/*
 * bsc1173663_cmd_parser.c
 *
 * Fix for CVE-2019-0155, bsc#1173663 (i915_cmd_parser.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)
#include "bsc1173663_common.h"
#include "../kallsyms_relocs.h"

/* klp-ccp: from drivers/gpu/drm/i915/i915_reg.h */
#define INSTR_CLIENT_SHIFT      29
#define INSTR_CLIENT_MASK       0xE0000000
#define   INSTR_MI_CLIENT       0x0
#define   INSTR_BC_CLIENT       0x2

#define MI_INSTR(opcode, flags) (((opcode) << 23) | (flags))

#define MI_NOOP			MI_INSTR(0, 0)
#define MI_USER_INTERRUPT	MI_INSTR(0x02, 0)
#define MI_WAIT_FOR_EVENT       MI_INSTR(0x03, 0)

#define MI_FLUSH		MI_INSTR(0x04, 0)

#define MI_REPORT_HEAD		MI_INSTR(0x07, 0)
#define MI_ARB_ON_OFF		MI_INSTR(0x08, 0)

#define MI_BATCH_BUFFER_END	MI_INSTR(0x0a, 0)
#define MI_SUSPEND_FLUSH	MI_INSTR(0x0b, 0)

#define MI_LOAD_SCAN_LINES_INCL MI_INSTR(0x12, 0)

#define MI_STORE_DWORD_IMM	MI_INSTR(0x20, 1)

#define MI_LOAD_REGISTER_IMM(x)	MI_INSTR(0x22, 2*(x)-1)

#define MI_STORE_REGISTER_MEM_GEN8   MI_INSTR(0x24, 2)

#define MI_FLUSH_DW		MI_INSTR(0x26, 1) /* for GEN6 */

#define MI_LOAD_REGISTER_MEM	   MI_INSTR(0x29, 1)

#define MI_LOAD_REGISTER_MEM_GEN8  MI_INSTR(0x29, 2)

#define   MI_BATCH_PPGTT_HSW		(1<<8)

#define MI_BATCH_BUFFER_START	MI_INSTR(0x31, 0)

#define MI_BATCH_BUFFER_START_GEN8	MI_INSTR(0x31, 1)

#define MI_ARB_CHECK            MI_INSTR(0x05, 0)

#define MI_LOAD_SCAN_LINES_EXCL MI_INSTR(0x13, 0)

#define MI_UPDATE_GTT           MI_INSTR(0x23, 0)

#define MI_LOAD_REGISTER_REG    MI_INSTR(0x2A, 0)

#define BCS_SWCTRL 0x22200

#define BCS_GPR(n)     (0x22600 + (n) * 8)
#define BCS_GPR_UDW(n) (0x22600 + (n) * 8 + 4)

#define OACONTROL 0x2360

#define RENDER_RING_BASE	0x02000
#define BSD_RING_BASE		0x04000

#define BLT_RING_BASE		0x22000

#define RING_TIMESTAMP(base)   ((base)+0x358)
#define RING_TIMESTAMP_UDW(base) ((base) + 0x358 + 4)

/* klp-ccp: from include/drm/drmP.h */
static void (*kgre_drm_clflush_virt_range)(void *addr, unsigned long length);

/* klp-ccp: from drivers/gpu/drm/i915/i915_drv.h */
#undef WARN_ON
#define WARN_ON(x) WARN((x), "WARN_ON(%s)", #x )

static int (*kgre_i915_gem_obj_prepare_shmem_read)(struct drm_i915_gem_object *obj,
				    int *needs_clflush);

static inline void i915_gem_object_unpin_pages(struct drm_i915_gem_object *obj)
{
	BUG_ON(obj->pages_pin_count == 0);
	obj->pages_pin_count--;
}

static int __must_check
(*kgre_i915_gem_object_set_to_cpu_domain)(struct drm_i915_gem_object *obj, bool write);

/* klp-ccp: from drivers/gpu/drm/i915/i915_cmd_parser.c */
#define STD_MI_OPCODE_MASK  0xFF800000

#define CMD(op, opm, f, lm, fl, ...)				\
	{							\
		.flags = (fl) | ((f) ? CMD_DESC_FIXED : 0),	\
		.cmd = { (op), (opm) },				\
		.length = { (lm) },				\
		__VA_ARGS__					\
	}

#define SMI STD_MI_OPCODE_MASK

#define F true
#define S CMD_DESC_SKIP

#define W CMD_DESC_REGISTER
#define B CMD_DESC_BITMASK

/* New. */
static const struct drm_i915_cmd_descriptor kgrp_gen9_blt_cmds[] = {
	CMD(  MI_NOOP,				SMI,	F,  1,	    S  ),
	CMD(  MI_USER_INTERRUPT,		SMI,	F,  1,	    S  ),
	CMD(  MI_WAIT_FOR_EVENT,		SMI,	F,  1,	    S  ),
	CMD(  MI_FLUSH,				SMI,	F,  1,	    S  ),
	CMD(  MI_ARB_CHECK,			SMI,	F,  1,	    S  ),
	CMD(  MI_REPORT_HEAD,			SMI,	F,  1,	    S  ),
	CMD(  MI_ARB_ON_OFF,			SMI,	F,  1,	    S  ),
	CMD(  MI_SUSPEND_FLUSH,			SMI,	F,  1,	    S  ),
	CMD(  MI_LOAD_SCAN_LINES_INCL,		SMI,   !F,  0x3F,   S  ),
	CMD(  MI_LOAD_SCAN_LINES_EXCL,		SMI,   !F,  0x3F,   S  ),
	CMD(  MI_STORE_DWORD_IMM,		SMI,   !F,  0x3FF,  S  ),
	CMD(  MI_LOAD_REGISTER_IMM(1),		SMI,   !F,  0xFF,   W,
	      .reg = { .offset = 1, .mask = 0x007FFFFC, .step = 2 }    ),
	CMD(  MI_UPDATE_GTT,			SMI,   !F,  0x3FF,  S  ),
	CMD(  MI_STORE_REGISTER_MEM_GEN8,	SMI,	F,  4,	    W,
	      .reg = { .offset = 1, .mask = 0x007FFFFC }	       ),
	CMD(  MI_FLUSH_DW,			SMI,   !F,  0x3F,   S  ),
	CMD(  MI_LOAD_REGISTER_MEM_GEN8,	SMI,	F,  4,	    W,
	      .reg = { .offset = 1, .mask = 0x007FFFFC }	       ),
	CMD(  MI_LOAD_REGISTER_REG,		SMI,	!F,  0xFF,  W,
	      .reg = { .offset = 1, .mask = 0x007FFFFC, .step = 1 }    ),

	 /*
	 * We allow BB_START but apply further checks. We just sanitize the
	 * basic fields here.
	 */
	CMD( MI_BATCH_BUFFER_START_GEN8,       SMI,    !F,  0xFF,  B,
	     .bits = {{
			.offset = 0,
			.mask = ~SMI,
			.expected = (MI_BATCH_PPGTT_HSW | 1),
	      }},                                            ),
};

#undef SMI

#undef F
#undef S

#undef W
#undef B

/* New. */
static const struct drm_i915_cmd_table kgrp_gen9_blt_cmd_table[] = {
	{ kgrp_gen9_blt_cmds, ARRAY_SIZE(kgrp_gen9_blt_cmds) },
};

struct drm_i915_reg_descriptor {
	u32 addr;
	u32 mask;
	u32 value;
};

#define REG32(address, ...)                             \
	{ .addr = address, __VA_ARGS__ }

#define REG64_IDX(_reg, idx) \
	{ .addr = _reg(idx) }, \
	{ .addr = _reg ## _UDW(idx) }

/* New. */
static const struct drm_i915_reg_descriptor kgrp_gen9_blt_regs[] = {
	REG64_IDX(RING_TIMESTAMP, RENDER_RING_BASE),
	REG64_IDX(RING_TIMESTAMP, BSD_RING_BASE),
	REG32(BCS_SWCTRL),
	REG64_IDX(RING_TIMESTAMP, BLT_RING_BASE),
	REG64_IDX(BCS_GPR, 0),
	REG64_IDX(BCS_GPR, 1),
	REG64_IDX(BCS_GPR, 2),
	REG64_IDX(BCS_GPR, 3),
	REG64_IDX(BCS_GPR, 4),
	REG64_IDX(BCS_GPR, 5),
	REG64_IDX(BCS_GPR, 6),
	REG64_IDX(BCS_GPR, 7),
	REG64_IDX(BCS_GPR, 8),
	REG64_IDX(BCS_GPR, 9),
	REG64_IDX(BCS_GPR, 10),
	REG64_IDX(BCS_GPR, 11),
	REG64_IDX(BCS_GPR, 12),
	REG64_IDX(BCS_GPR, 13),
	REG64_IDX(BCS_GPR, 14),
	REG64_IDX(BCS_GPR, 15),
};

#undef REG64_IDX
#undef REG32

/* New. */
static u32 kgrp_gen9_blt_get_cmd_length_mask(u32 cmd_header)
{
	u32 client = (cmd_header & INSTR_CLIENT_MASK) >> INSTR_CLIENT_SHIFT;

	if (client == INSTR_MI_CLIENT || client == INSTR_BC_CLIENT)
		return 0xFF;

	KGRR_DRM_DEBUG_DRIVER("CMD: Abnormal blt cmd length! 0x%08X\n", cmd_header);
	return 0;
}

struct cmd_node {
	const struct drm_i915_cmd_descriptor *desc;
	struct hlist_node node;
};

#define CMD_HASH_MASK STD_MI_OPCODE_MASK

/*
 * Copy of the original init_hash_table with slightly modified
 * signature for initializing the static kgrp_gen9_blt_cmd_hash.
 */
static int kgrp_init_hash_table(DECLARE_HASHTABLE((*cmd_hash), I915_CMD_HASH_ORDER),
			   const struct drm_i915_cmd_table *cmd_tables,
			   int cmd_table_count)
{
	int i, j;

	hash_init((*cmd_hash));

	for (i = 0; i < cmd_table_count; i++) {
		const struct drm_i915_cmd_table *table = &cmd_tables[i];

		for (j = 0; j < table->count; j++) {
			const struct drm_i915_cmd_descriptor *desc =
				&table->table[j];
			struct cmd_node *desc_node =
				kmalloc(sizeof(*desc_node), GFP_KERNEL);

			if (!desc_node)
				return -ENOMEM;

			desc_node->desc = desc;
			hash_add((*cmd_hash), &desc_node->node,
				 desc->cmd.value & CMD_HASH_MASK);
		}
	}

	return 0;
}

/*
 * Copy of the original init_hash_table with slightly modified
 * signature for destroying the static kgrp_gen9_blt_cmd_hash.
 */
static void kgrp_fini_hash_table(DECLARE_HASHTABLE((*cmd_hash), I915_CMD_HASH_ORDER))
{
	struct hlist_node *tmp;
	struct cmd_node *desc_node;
	int i;

	hash_for_each_safe((*cmd_hash), i, tmp, desc_node, node) {
		hash_del(&desc_node->node);
		kfree(desc_node);
	}
}

/* New. */
static DECLARE_HASHTABLE(kgrp_gen9_blt_cmd_hash, I915_CMD_HASH_ORDER);

static const struct drm_i915_cmd_descriptor*
kgrp_find_cmd_in_table(struct intel_engine_cs *ring,
		  u32 cmd_header)
{
	struct cmd_node *desc_node;
	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +5 lines
	 * Special case on gen9 blt engine and use
	 * the static hashtable in this case.
	 */
	DECLARE_HASHTABLE((*cmd_hash), I915_CMD_HASH_ORDER) = &ring->cmd_hash;
	if (kgrp_is_gen9_blt(ring))
		cmd_hash = &kgrp_gen9_blt_cmd_hash;

	hash_for_each_possible((*cmd_hash), desc_node, node,
			       cmd_header & CMD_HASH_MASK) {
		const struct drm_i915_cmd_descriptor *desc = desc_node->desc;
		u32 masked_cmd = desc->cmd.mask & cmd_header;
		u32 masked_value = desc->cmd.value & desc->cmd.mask;

		if (masked_cmd == masked_value)
			return desc;
	}

	return NULL;
}

static const struct drm_i915_cmd_descriptor*
kgrp_find_cmd(struct intel_engine_cs *ring,
	 u32 cmd_header,
	 struct drm_i915_cmd_descriptor *default_desc)
{
	const struct drm_i915_cmd_descriptor *desc;
	u32 mask;

	desc = kgrp_find_cmd_in_table(ring, cmd_header);
	if (desc)
		return desc;

	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +4 lines
	 * Special case on gen9 blt engine and use
	 * kgrp_gen9_blt_get_cmd_length_mask() in this case.
	 */
	if (!kgrp_is_gen9_blt(ring))
		mask = ring->get_cmd_length_mask(cmd_header);
	else
		mask = kgrp_gen9_blt_get_cmd_length_mask(cmd_header);
	if (!mask)
		return NULL;

	BUG_ON(!default_desc);
	default_desc->flags = CMD_DESC_SKIP;
	default_desc->length.mask = mask;

	return default_desc;
}

static const struct drm_i915_reg_descriptor *
find_reg(const struct drm_i915_reg_descriptor *table,
	 int count, u32 addr)
{
	if (table) {
		int i;

		for (i = 0; i < count; i++) {
			if (table[i].addr == addr)
				return &table[i];
		}
	}

	return NULL;
}

static u32 *kgrr_vmap_batch(struct drm_i915_gem_object *obj,
		       unsigned start, unsigned len)
{
	int i;
	void *addr = NULL;
	struct sg_page_iter sg_iter;
	int first_page = start >> PAGE_SHIFT;
	int last_page = (len + start + 4095) >> PAGE_SHIFT;
	int npages = last_page - first_page;
	struct page **pages;

	pages = drm_malloc_ab(npages, sizeof(*pages));
	if (pages == NULL) {
		KGRR_DRM_DEBUG_DRIVER("Failed to get space for pages\n");
		goto finish;
	}

	i = 0;
	for_each_sg_page(obj->pages->sgl, &sg_iter, obj->pages->nents, first_page) {
		pages[i++] = sg_page_iter_page(&sg_iter);
		if (i == npages)
			break;
	}

	addr = vmap(pages, i, 0, PAGE_KERNEL);
	if (addr == NULL) {
		KGRR_DRM_DEBUG_DRIVER("Failed to vmap pages\n");
		goto finish;
	}

finish:
	if (pages)
		drm_free_large(pages);
	return (u32*)addr;
}

static u32 *kgrr_copy_batch(struct drm_i915_gem_object *dest_obj,
		       struct drm_i915_gem_object *src_obj,
		       u32 batch_start_offset,
		       u32 batch_len)
{
	int needs_clflush = 0;
	void *src_base, *src;
	void *dst = NULL;
	int ret;

	if (batch_len > dest_obj->base.size ||
	    batch_len + batch_start_offset > src_obj->base.size)
		return ERR_PTR(-E2BIG);

	if (WARN_ON(dest_obj->pages_pin_count == 0))
		return ERR_PTR(-ENODEV);

	ret = (*kgre_i915_gem_obj_prepare_shmem_read)(src_obj, &needs_clflush);
	if (ret) {
		KGRR_DRM_DEBUG_DRIVER("CMD: failed to prepare shadow batch\n");
		return ERR_PTR(ret);
	}

	src_base = kgrr_vmap_batch(src_obj, batch_start_offset, batch_len);
	if (!src_base) {
		KGRR_DRM_DEBUG_DRIVER("CMD: Failed to vmap batch\n");
		ret = -ENOMEM;
		goto unpin_src;
	}

	ret = (*kgre_i915_gem_object_set_to_cpu_domain)(dest_obj, true);
	if (ret) {
		KGRR_DRM_DEBUG_DRIVER("CMD: Failed to set shadow batch to CPU\n");
		goto unmap_src;
	}

	dst = kgrr_vmap_batch(dest_obj, 0, batch_len);
	if (!dst) {
		KGRR_DRM_DEBUG_DRIVER("CMD: Failed to vmap shadow batch\n");
		ret = -ENOMEM;
		goto unmap_src;
	}

	src = src_base + offset_in_page(batch_start_offset);
	if (needs_clflush)
		(*kgre_drm_clflush_virt_range)(src, batch_len);

	memcpy(dst, src, batch_len);

unmap_src:
	vunmap(src_base);
unpin_src:
	i915_gem_object_unpin_pages(src_obj);

	return ret ? ERR_PTR(ret) : dst;
}

/*
 * Don't bother with adding shadow variables to struct intel_context
 * and simply allocate a whitelist for each i915_parse_cmds()
 * invocation.
 */
struct kgrp_jump_whitelist
{
	unsigned long *jump_whitelist;
	uint32_t jump_whitelist_cmds;
};

static int kgrp_check_bbstart(struct intel_context *ctx,
			 struct kgrp_jump_whitelist *jw,
			 u32 *cmd, u64 offset, u32 length,
			 u32 batch_len,
			 u64 batch_start,
			 u64 shadow_batch_start)
{

	u64 jump_offset, jump_target;
	u32 target_cmd_offset, target_cmd_index;

	/* For igt compatibility on older platforms */
	if (KGRP_CMDPARSER_USES_GGTT(ctx->i915)) {
		KGRR_DRM_DEBUG("CMD: Rejecting BB_START for ggtt based submission\n");
		return -EACCES;
	}

	if (length != 3) {
		KGRR_DRM_DEBUG("CMD: Recursive BB_START with bad length(%u)\n",
				 length);
		return -EINVAL;
	}

	jump_target = *(u64*)(cmd+1);
	jump_offset = jump_target - batch_start;

	/*
	 * Any underflow of jump_target is guaranteed to be outside the range
	 * of a u32, so >= test catches both too large and too small
	*/
	if (jump_offset >= batch_len) {
		KGRR_DRM_DEBUG("CMD: BB_START to 0x%llx jumps out of BB\n",
			  jump_target);
		return -EINVAL;
	}

	/*
	 * This cannot overflow a u32 because we already checked jump_offset
	 * is within the BB, and the batch_len is a u32
	 */
	target_cmd_offset = lower_32_bits(jump_offset);
	target_cmd_index = target_cmd_offset / sizeof(u32);

	*(u64*)(cmd + 1) = shadow_batch_start + target_cmd_offset;

	if (target_cmd_index == offset)
		return 0;

	if (jw->jump_whitelist_cmds <= target_cmd_index) {
		KGRR_DRM_DEBUG("CMD: Rejecting BB_START - truncated whitelist array\n");
		return -EINVAL;
	} else if (!test_bit(target_cmd_index, jw->jump_whitelist)) {
		KGRR_DRM_DEBUG("CMD: BB_START to 0x%llx not a previously executed cmd\n",
			  jump_target);
		return -EINVAL;
	}

	return 0;
}

static void kgrp_init_whitelist(struct intel_context *ctx,
				struct kgrp_jump_whitelist *jw, u32 batch_len)
{
	const u32 batch_cmds = DIV_ROUND_UP(batch_len, sizeof(u32));
	const u32 exact_size = BITS_TO_LONGS(batch_cmds);
	unsigned long *next_whitelist;

	memset(jw, 0, sizeof(*jw));

	if (KGRP_CMDPARSER_USES_GGTT(ctx->i915))
		return;

	next_whitelist = kcalloc(exact_size, sizeof(long), GFP_KERNEL);
	if (next_whitelist) {
		jw->jump_whitelist = next_whitelist;
		jw->jump_whitelist_cmds =
			exact_size * BITS_PER_BYTE * sizeof(long);
		return;
	}

	KGRR_DRM_DEBUG("CMD: Failed to extend whitelist. BB_START may be disallowed\n");

	return;
}

static bool kgrp_check_cmd(const struct intel_engine_cs *ring,
		      const struct drm_i915_cmd_descriptor *desc,
		      const u32 *cmd, u32 length,
		      /*
		       * Fix CVE-2019-0155
		       *  -1 line
		       */
		      bool *oacontrol_set)
{
	if (desc->flags & CMD_DESC_REJECT) {
		KGRR_DRM_DEBUG_DRIVER("CMD: Rejected command: 0x%08X\n", *cmd);
		return false;
	}

	/*
	 * Fix CVE-2019-0155
	 *  -5 lines
	 * Don't reject CMD_DESC_MASTER, c.f. upstream commit
	 * 66d8aba1cd6d ("drm/i915: Remove Master tables from cmdparser").
	 */

	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +2 lines
	 * Treat CMD_DESC_MASTER as CMD_DESC_REGISTER, c.f. upstream commit
	 * 66d8aba1cd6d ("drm/i915: Remove Master tables from cmdparser").
	 */
	if ((desc->flags & CMD_DESC_REGISTER) ||
	    (desc->flags & CMD_DESC_MASTER)) {
		/*
		 * Get the distance between individual register offset
		 * fields if the command can perform more than one
		 * access at a time.
		 */
		const u32 step = desc->reg.step ? desc->reg.step : length;
		u32 offset;

		for (offset = desc->reg.offset; offset < length;
		     offset += step) {
			const u32 reg_addr = cmd[offset] & desc->reg.mask;
			/*
			 * Fix CVE-2019-0155
			 *  -3 lines, +7 lines
			 * Special case on gen9 blt engine and use
			 * the static register table in this case.
			 */
			const struct drm_i915_reg_descriptor *reg =
				(!kgrp_is_gen9_blt(ring) ?
				  find_reg(ring->reg_table, ring->reg_count,
					   reg_addr) :
				  find_reg(kgrp_gen9_blt_regs,
					   ARRAY_SIZE(kgrp_gen9_blt_regs),
					   reg_addr));

			/*
			 * Fix CVE-2019-0155
			 *  -4 lines
			 * Never search master reg tables, c.f. upstream commit
			 * 66d8aba1cd6d ("drm/i915: Remove Master tables from
			 * cmdparser").
			 */

			if (!reg) {
				KGRR_DRM_DEBUG_DRIVER("CMD: Rejected register 0x%08X in command: 0x%08X (ring=%d)\n",
						 reg_addr, *cmd, ring->id);
				return false;
			}

			/*
			 * OACONTROL requires some special handling for
			 * writes. We want to make sure that any batch which
			 * enables OA also disables it before the end of the
			 * batch. The goal is to prevent one process from
			 * snooping on the perf data from another process. To do
			 * that, we need to check the value that will be written
			 * to the register. Hence, limit OACONTROL writes to
			 * only MI_LOAD_REGISTER_IMM commands.
			 */
			if (reg_addr == OACONTROL) {
				/*
				 * Fix CVE-2019-0155
				 *  -1 line, +2 lines
				 */
				if (desc->cmd.value == MI_LOAD_REGISTER_MEM ||
				    desc->cmd.value == MI_LOAD_REGISTER_MEM_GEN8) {
					KGRR_DRM_DEBUG_DRIVER("CMD: Rejected LRM to OACONTROL\n");
					return false;
				}

				if (desc->cmd.value == MI_LOAD_REGISTER_IMM(1))
					*oacontrol_set = (cmd[offset + 1] != 0);
			}

			/*
			 * Check the value written to the register against the
			 * allowed mask/value pair given in the whitelist entry.
			 */
			if (reg->mask) {
				/*
				 * Fix CVE-2019-0155
				 *  -1 line, +2 lines
				 */
				if (desc->cmd.value == MI_LOAD_REGISTER_MEM ||
				    desc->cmd.value == MI_LOAD_REGISTER_MEM_GEN8) {
					KGRR_DRM_DEBUG_DRIVER("CMD: Rejected LRM to masked register 0x%08X\n",
							 reg_addr);
					return false;
				}

				if (desc->cmd.value == MI_LOAD_REGISTER_IMM(1) &&
				    (offset + 2 > length ||
				     (cmd[offset + 1] & reg->mask) != reg->value)) {
					KGRR_DRM_DEBUG_DRIVER("CMD: Rejected LRI to masked register 0x%08X\n",
							 reg_addr);
					return false;
				}
			}
		}
	}

	if (desc->flags & CMD_DESC_BITMASK) {
		int i;

		for (i = 0; i < MAX_CMD_DESC_BITMASKS; i++) {
			u32 dword;

			if (desc->bits[i].mask == 0)
				break;

			if (desc->bits[i].condition_mask != 0) {
				u32 offset =
					desc->bits[i].condition_offset;
				u32 condition = cmd[offset] &
					desc->bits[i].condition_mask;

				if (condition == 0)
					continue;
			}

			if (desc->bits[i].offset >= length) {
				KGRR_DRM_DEBUG_DRIVER("CMD: Rejected command 0x%08X, too short to check bitmask (%s)\n",
						 *cmd, ring->name);
				return false;
			}

			dword = cmd[desc->bits[i].offset] &
				desc->bits[i].mask;

			if (dword != desc->bits[i].expected) {
				KGRR_DRM_DEBUG_DRIVER("CMD: Rejected command 0x%08X for bitmask 0x%08X (exp=0x%08X act=0x%08X) (ring=%d)\n",
						 *cmd,
						 desc->bits[i].mask,
						 desc->bits[i].expected,
						 dword, ring->id);
				return false;
			}
		}
	}

	return true;
}

#define LENGTH_BIAS 2

/*
 * Fix CVE-2019-0155
 *  -6 lines, +8 lines
 */
int kgrp_i915_parse_cmds(struct intel_context *ctx,
		    struct intel_engine_cs *ring,
		    struct drm_i915_gem_object *batch_obj,
		    u64 user_batch_start,
		    u32 batch_start_offset,
		    u32 batch_len,
		    struct drm_i915_gem_object *shadow_batch_obj,
		    u64 shadow_batch_start)
{
	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +1 line
	 */
	u32 *cmd, *batch_base, *batch_end, offset = 0;
	struct drm_i915_cmd_descriptor default_desc = { 0 };
	bool oacontrol_set = false; /* OACONTROL tracking. See check_cmd() */
	int ret = 0;
	/*
	 * Fix CVE-2019-0155
	 *  +1 line
	 */
	struct kgrp_jump_whitelist jw;

	batch_base = kgrr_copy_batch(shadow_batch_obj, batch_obj,
				batch_start_offset, batch_len);
	if (IS_ERR(batch_base)) {
		KGRR_DRM_DEBUG_DRIVER("CMD: Failed to copy batch\n");
		return PTR_ERR(batch_base);
	}

	/*
	 * Fix CVE-2019-0155
	 *  +1 line
	 */
	kgrp_init_whitelist(ctx, &jw, batch_len);

	/*
	 * We use the batch length as size because the shadow object is as
	 * large or larger and copy_batch() will write MI_NOPs to the extra
	 * space. Parsing should be faster in some cases this way.
	 */
	batch_end = batch_base + (batch_len / sizeof(*batch_end));

	cmd = batch_base;
	while (cmd < batch_end) {
		const struct drm_i915_cmd_descriptor *desc;
		u32 length;

		if (*cmd == MI_BATCH_BUFFER_END)
			break;

		desc = kgrp_find_cmd(ring, *cmd, &default_desc);
		if (!desc) {
			KGRR_DRM_DEBUG_DRIVER("CMD: Unrecognized command: 0x%08X\n",
					 *cmd);
			ret = -EINVAL;
			break;
		}

		/*
		 * Fix CVE-2019-0155
		 *  -9 lines
		 */

		if (desc->flags & CMD_DESC_FIXED)
			length = desc->length.fixed;
		else
			length = ((*cmd & desc->length.mask) + LENGTH_BIAS);

		if ((batch_end - cmd) < length) {
			KGRR_DRM_DEBUG_DRIVER("CMD: Command length exceeds batch length: 0x%08X length=%u batchlen=%td\n",
					 *cmd,
					 length,
					 batch_end - cmd);
			ret = -EINVAL;
			break;
		}

		/*
		 * Fix CVE-2019-0155
		 *  -2 lines, +1 line
		 */
		if (!kgrp_check_cmd(ring, desc, cmd, length, &oacontrol_set)) {
			/*
			 * Fix CVE-2019-0155
			 *  -1 line, +1 line
			 */
			ret = KGRP_CMDPARSER_USES_GGTT(ring->dev) ? -EINVAL : -EACCES;
			break;
		}

		/*
		 * Fix CVE-2019-0155
		 *  +11 lines
		 */
		if (desc->cmd.value == MI_BATCH_BUFFER_START ||
		    desc->cmd.value == MI_BATCH_BUFFER_START_GEN8) {
			ret = kgrp_check_bbstart(ctx, &jw, cmd, offset, length,
					    batch_len, user_batch_start,
					    shadow_batch_start);
			break;
		}

		if (jw.jump_whitelist_cmds > offset)
			set_bit(offset, jw.jump_whitelist);

		cmd += length;
		/*
		 * Fix CVE-2019-0155
		 *  +1 line
		 */
		offset += length;
	}

	if (oacontrol_set) {
		KGRR_DRM_DEBUG_DRIVER("CMD: batch set OACONTROL but did not clear it\n");
		ret = -EINVAL;
	}

	if (cmd >= batch_end) {
		KGRR_DRM_DEBUG_DRIVER("CMD: Got to the end of the buffer w/o a BBE cmd!\n");
		ret = -EINVAL;
	}

	/*
	 * Fix CVE-2019-0155
	 *  +1 line
	 */
	kfree(jw.jump_whitelist);
	vunmap(batch_base);

	return ret;
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "i915_gem_object_set_to_cpu_domain",
	  (void *)&kgre_i915_gem_object_set_to_cpu_domain, "i915" },
	{ "i915_gem_obj_prepare_shmem_read",
	  (void *)&kgre_i915_gem_obj_prepare_shmem_read, "i915" },
	{ "drm_clflush_virt_range", (void *)&kgre_drm_clflush_virt_range,
	  "drm" },
};

int kgr_bsc1173663_cmd_parser_resolve_kallsyms(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}

int kgr_bsc1173663_cmd_parser_init(void)
{
	return kgrp_init_hash_table(&kgrp_gen9_blt_cmd_hash,
				    kgrp_gen9_blt_cmd_table,
				    ARRAY_SIZE(kgrp_gen9_blt_cmd_table));
}

void kgr_bsc1173663_cmd_parser_cleanup(void)
{
	kgrp_fini_hash_table(&kgrp_gen9_blt_cmd_hash);
}

#endif
