/*
 * bsc1173663_gem.c
 *
 * Fix for CVE-2019-0155, bsc#1173663 (i915_gem.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)
#include "bsc1173663_common.h"
#include "../kallsyms_relocs.h"

/* klp-ccp: from include/drm/drm_mm.h */
static int (*kgre_drm_mm_insert_node_in_range_generic)(struct drm_mm *mm,
					struct drm_mm_node *node,
					u64 size,
					unsigned alignment,
					unsigned long color,
					u64 start,
					u64 end,
					enum drm_mm_search_flags sflags,
					enum drm_mm_allocator_flags aflags);

static void (*kgre_drm_mm_remove_node)(struct drm_mm_node *node);

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.h */
static const struct i915_ggtt_view (*kgre_i915_ggtt_view_normal);

static size_t
(*kgre_i915_ggtt_view_size)(struct drm_i915_gem_object *obj,
		    const struct i915_ggtt_view *view);

/* klp-ccp: from drivers/gpu/drm/i915/i915_drv.h */
#undef WARN_ON
#define WARN_ON(x) WARN((x), "WARN_ON(%s)", #x )

/* klp-ccp: from drivers/gpu/drm/i915/i915_trace.h */
KGRR_TRACE_EVENT(i915_vma_bind,
	    TP_PROTO(struct i915_vma *vma, unsigned flags),
	    TP_ARGS(vma, flags)
);

/* klp-ccp: from drivers/gpu/drm/i915/i915_drv.h */
static void (*kgre_i915_gem_vma_destroy)(struct i915_vma *vma);

int __must_check
kgrp_i915_gem_object_pin(struct drm_i915_gem_object *obj,
		    struct i915_address_space *vm,
		    uint32_t alignment,
		    uint64_t flags);

static void (*kgre___i915_vma_set_map_and_fenceable)(struct i915_vma *vma);
static int __must_check (*kgre_i915_vma_unbind)(struct i915_vma *vma);

static int __must_check (*kgre_i915_gem_object_get_pages)(struct drm_i915_gem_object *obj);

static inline void i915_gem_object_pin_pages(struct drm_i915_gem_object *obj)
{
	BUG_ON(obj->pages == NULL);
	obj->pages_pin_count++;
}
static inline void i915_gem_object_unpin_pages(struct drm_i915_gem_object *obj)
{
	BUG_ON(obj->pages_pin_count == 0);
	obj->pages_pin_count--;
}

static uint32_t
(*kgre_i915_gem_get_gtt_size)(struct drm_device *dev, uint32_t size, int tiling_mode);
static uint32_t
(*kgre_i915_gem_get_gtt_alignment)(struct drm_device *dev, uint32_t size,
			    int tiling_mode, bool fenced);

static struct i915_vma *
(*kgre_i915_gem_obj_to_vma)(struct drm_i915_gem_object *obj,
		    struct i915_address_space *vm);
static struct i915_vma *
(*kgre_i915_gem_obj_to_ggtt_view)(struct drm_i915_gem_object *obj,
			  const struct i915_ggtt_view *view);

static struct i915_vma *
(*kgre_i915_gem_obj_lookup_or_create_vma)(struct drm_i915_gem_object *obj,
				  struct i915_address_space *vm);
static struct i915_vma *
(*kgre_i915_gem_obj_lookup_or_create_ggtt_vma)(struct drm_i915_gem_object *obj,
				       const struct i915_ggtt_view *view);

static int __must_check (*kgre_i915_gem_evict_something)(struct drm_device *dev,
					  struct i915_address_space *vm,
					  int min_size,
					  unsigned alignment,
					  unsigned cache_level,
					  unsigned long start,
					  unsigned long end,
					  unsigned flags);

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem.c */
static uint32_t
(*kgre_i915_gem_get_gtt_size)(struct drm_device *dev, uint32_t size, int tiling_mode);

static uint32_t
(*kgre_i915_gem_get_gtt_alignment)(struct drm_device *dev, uint32_t size,
			   int tiling_mode, bool fenced);

static int
(*kgre_i915_gem_object_get_pages)(struct drm_i915_gem_object *obj);

static int (*kgre_i915_vma_unbind)(struct i915_vma *vma);

static bool i915_gem_valid_gtt_space(struct i915_vma *vma,
				     unsigned long cache_level)
{
	struct drm_mm_node *gtt_space = &vma->node;
	struct drm_mm_node *other;

	/*
	 * On some machines we have to be careful when putting differing types
	 * of snoopable memory together to avoid the prefetcher crossing memory
	 * domains and dying. During vm initialisation, we decide whether or not
	 * these constraints apply and set the drm_mm.color_adjust
	 * appropriately.
	 */
	if (vma->vm->mm.color_adjust == NULL)
		return true;

	if (!drm_mm_node_allocated(gtt_space))
		return true;

	if (list_empty(&gtt_space->node_list))
		return true;

	other = list_entry(gtt_space->node_list.prev, struct drm_mm_node, node_list);
	if (other->allocated && !other->hole_follows && other->color != cache_level)
		return false;

	other = list_entry(gtt_space->node_list.next, struct drm_mm_node, node_list);
	if (other->allocated && !gtt_space->hole_follows && other->color != cache_level)
		return false;

	return true;
}

static struct i915_vma *
kgrp_i915_gem_object_bind_to_vm(struct drm_i915_gem_object *obj,
			   struct i915_address_space *vm,
			   const struct i915_ggtt_view *ggtt_view,
			   unsigned alignment,
			   uint64_t flags)
{
	struct drm_device *dev = obj->base.dev;
	struct drm_i915_private *dev_priv = dev->dev_private;
	u32 fence_alignment, unfenced_alignment;
	u32 search_flag, alloc_flag;
	u64 start, end;
	u64 size, fence_size;
	struct i915_vma *vma;
	int ret;

	if (i915_is_ggtt(vm)) {
		u32 view_size;

		if (WARN_ON(!ggtt_view))
			return ERR_PTR(-EINVAL);

		view_size = (*kgre_i915_ggtt_view_size)(obj, ggtt_view);

		fence_size = (*kgre_i915_gem_get_gtt_size)(dev,
						   view_size,
						   obj->tiling_mode);
		fence_alignment = (*kgre_i915_gem_get_gtt_alignment)(dev,
							     view_size,
							     obj->tiling_mode,
							     true);
		unfenced_alignment = (*kgre_i915_gem_get_gtt_alignment)(dev,
								view_size,
								obj->tiling_mode,
								false);
		size = flags & PIN_MAPPABLE ? fence_size : view_size;
	} else {
		fence_size = (*kgre_i915_gem_get_gtt_size)(dev,
						   obj->base.size,
						   obj->tiling_mode);
		fence_alignment = (*kgre_i915_gem_get_gtt_alignment)(dev,
							     obj->base.size,
							     obj->tiling_mode,
							     true);
		unfenced_alignment =
			(*kgre_i915_gem_get_gtt_alignment)(dev,
						   obj->base.size,
						   obj->tiling_mode,
						   false);
		size = flags & PIN_MAPPABLE ? fence_size : obj->base.size;
	}

	start = flags & PIN_OFFSET_BIAS ? flags & PIN_OFFSET_MASK : 0;
	end = vm->total;
	if (flags & PIN_MAPPABLE)
		end = min_t(u64, end, dev_priv->gtt.mappable_end);
	if (flags & PIN_ZONE_4G)
		end = min_t(u64, end, (1ULL << 32));

	if (alignment == 0)
		alignment = flags & PIN_MAPPABLE ? fence_alignment :
						unfenced_alignment;
	if (flags & PIN_MAPPABLE && alignment & (fence_alignment - 1)) {
		KGRR_DRM_DEBUG("Invalid object (view type=%u) alignment requested %u\n",
			  ggtt_view ? ggtt_view->type : 0,
			  alignment);
		return ERR_PTR(-EINVAL);
	}

	/* If binding the object/GGTT view requires more space than the entire
	 * aperture has, reject it early before evicting everything in a vain
	 * attempt to find space.
	 */
	if (size > end) {
		KGRR_DRM_DEBUG("Attempting to bind an object (view type=%u) larger than the aperture: size=%llu > %s aperture=%llu\n",
			  ggtt_view ? ggtt_view->type : 0,
			  size,
			  flags & PIN_MAPPABLE ? "mappable" : "total",
			  end);
		return ERR_PTR(-E2BIG);
	}

	ret = (*kgre_i915_gem_object_get_pages)(obj);
	if (ret)
		return ERR_PTR(ret);

	i915_gem_object_pin_pages(obj);

	vma = ggtt_view ? (*kgre_i915_gem_obj_lookup_or_create_ggtt_vma)(obj, ggtt_view) :
			  (*kgre_i915_gem_obj_lookup_or_create_vma)(obj, vm);

	if (IS_ERR(vma))
		goto err_unpin;

	if (flags & PIN_HIGH) {
		search_flag = DRM_MM_SEARCH_BELOW;
		alloc_flag = DRM_MM_CREATE_TOP;
	} else {
		search_flag = DRM_MM_SEARCH_DEFAULT;
		alloc_flag = DRM_MM_CREATE_DEFAULT;
	}

search_free:
	ret = (*kgre_drm_mm_insert_node_in_range_generic)(&vm->mm, &vma->node,
						  size, alignment,
						  obj->cache_level,
						  start, end,
						  search_flag,
						  alloc_flag);
	if (ret) {
		ret = (*kgre_i915_gem_evict_something)(dev, vm, size, alignment,
					       obj->cache_level,
					       start, end,
					       flags);
		if (ret == 0)
			goto search_free;

		goto err_free_vma;
	}
	if (WARN_ON(!i915_gem_valid_gtt_space(vma, obj->cache_level))) {
		ret = -EINVAL;
		goto err_remove_node;
	}

	kgrr_trace_i915_vma_bind(vma, flags);
	ret = kgrp_i915_vma_bind(vma, obj->cache_level, flags);
	if (ret)
		goto err_remove_node;

	list_move_tail(&obj->global_list, &dev_priv->mm.bound_list);
	list_add_tail(&vma->mm_list, &vm->inactive_list);

	return vma;

err_remove_node:
	(*kgre_drm_mm_remove_node)(&vma->node);
err_free_vma:
	(*kgre_i915_gem_vma_destroy)(vma);
	vma = ERR_PTR(ret);
err_unpin:
	i915_gem_object_unpin_pages(obj);
	return vma;
}

static bool
i915_vma_misplaced(struct i915_vma *vma, uint32_t alignment, uint64_t flags)
{
	struct drm_i915_gem_object *obj = vma->obj;

	if (alignment &&
	    vma->node.start & (alignment - 1))
		return true;

	if (flags & PIN_MAPPABLE && !obj->map_and_fenceable)
		return true;

	if (flags & PIN_OFFSET_BIAS &&
	    vma->node.start < (flags & PIN_OFFSET_MASK))
		return true;

	return false;
}

static void (*kgre___i915_vma_set_map_and_fenceable)(struct i915_vma *vma);

static int
kgrp_i915_gem_object_do_pin(struct drm_i915_gem_object *obj,
		       struct i915_address_space *vm,
		       const struct i915_ggtt_view *ggtt_view,
		       uint32_t alignment,
		       uint64_t flags)
{
	struct drm_i915_private *dev_priv = obj->base.dev->dev_private;
	struct i915_vma *vma;
	unsigned bound;
	int ret;

	if (WARN_ON(vm == &dev_priv->mm.aliasing_ppgtt->base))
		return -ENODEV;

	if (WARN_ON(flags & (PIN_GLOBAL | PIN_MAPPABLE) && !i915_is_ggtt(vm)))
		return -EINVAL;

	if (WARN_ON((flags & (PIN_MAPPABLE | PIN_GLOBAL)) == PIN_MAPPABLE))
		return -EINVAL;

	if (WARN_ON(i915_is_ggtt(vm) != !!ggtt_view))
		return -EINVAL;

	vma = ggtt_view ? (*kgre_i915_gem_obj_to_ggtt_view)(obj, ggtt_view) :
			  (*kgre_i915_gem_obj_to_vma)(obj, vm);

	if (IS_ERR(vma))
		return PTR_ERR(vma);

	if (vma) {
		if (WARN_ON(vma->pin_count == DRM_I915_GEM_OBJECT_MAX_PIN_COUNT))
			return -EBUSY;

		if (i915_vma_misplaced(vma, alignment, flags)) {
			WARN(vma->pin_count,
			     "bo is already pinned in %s with incorrect alignment:"
			     " offset=%08x %08x, req.alignment=%x, req.map_and_fenceable=%d,"
			     " obj->map_and_fenceable=%d\n",
			     ggtt_view ? "ggtt" : "ppgtt",
			     upper_32_bits(vma->node.start),
			     lower_32_bits(vma->node.start),
			     alignment,
			     !!(flags & PIN_MAPPABLE),
			     obj->map_and_fenceable);
			ret = (*kgre_i915_vma_unbind)(vma);
			if (ret)
				return ret;

			vma = NULL;
		}
	}

	bound = vma ? vma->bound : 0;
	if (vma == NULL || !drm_mm_node_allocated(&vma->node)) {
		vma = kgrp_i915_gem_object_bind_to_vm(obj, vm, ggtt_view, alignment,
						 flags);
		if (IS_ERR(vma))
			return PTR_ERR(vma);
	} else {
		ret = kgrp_i915_vma_bind(vma, obj->cache_level, flags);
		if (ret)
			return ret;
	}

	if (ggtt_view && ggtt_view->type == I915_GGTT_VIEW_NORMAL &&
	    (bound ^ vma->bound) & GLOBAL_BIND) {
		(*kgre___i915_vma_set_map_and_fenceable)(vma);
		WARN_ON(flags & PIN_MAPPABLE && !obj->map_and_fenceable);
	}

	vma->pin_count++;
	return 0;
}

/*
 * Patched to (statically) call kgrp_i915_vma_bind() supporting RO
 * PTEs on gen8+ down the callchain.
*/
int
kgrp_i915_gem_object_pin(struct drm_i915_gem_object *obj,
		    struct i915_address_space *vm,
		    uint32_t alignment,
		    uint64_t flags)
{
	return kgrp_i915_gem_object_do_pin(obj, vm,
				      i915_is_ggtt(vm) ? &(*kgre_i915_ggtt_view_normal) : NULL,
				      alignment, flags);
}


static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "i915_ggtt_view_normal", (void *)&kgre_i915_ggtt_view_normal,
	  "i915" },
	{ "__tracepoint_i915_vma_bind",
	  (void *)&kgre___tracepoint_i915_vma_bind, "i915" },
	{ "i915_ggtt_view_size", (void *)&kgre_i915_ggtt_view_size,
	  "i915" },
	{ "i915_gem_evict_something", (void *)&kgre_i915_gem_evict_something,
	  "i915" },
	{ "i915_gem_vma_destroy", (void *)&kgre_i915_gem_vma_destroy,
	  "i915" },
	{ "i915_gem_get_gtt_size", (void *)&kgre_i915_gem_get_gtt_size,
	  "i915" },
	{ "i915_gem_object_get_pages", (void *)&kgre_i915_gem_object_get_pages,
	  "i915" },
	{ "i915_gem_get_gtt_alignment",
	  (void *)&kgre_i915_gem_get_gtt_alignment, "i915" },
	{ "__i915_vma_set_map_and_fenceable",
	  (void *)&kgre___i915_vma_set_map_and_fenceable, "i915" },
	{ "i915_vma_unbind", (void *)&kgre_i915_vma_unbind, "i915" },
	{ "i915_gem_obj_to_vma", (void *)&kgre_i915_gem_obj_to_vma, "i915" },
	{ "i915_gem_obj_to_ggtt_view", (void *)&kgre_i915_gem_obj_to_ggtt_view,
	  "i915" },
	{ "i915_gem_obj_lookup_or_create_vma",
	  (void *)&kgre_i915_gem_obj_lookup_or_create_vma, "i915" },
	{ "i915_gem_obj_lookup_or_create_ggtt_vma",
	  (void *)&kgre_i915_gem_obj_lookup_or_create_ggtt_vma, "i915" },
	{ "drm_mm_insert_node_in_range_generic",
	  (void *)&kgre_drm_mm_insert_node_in_range_generic, "drm" },
	{ "drm_mm_remove_node", (void *)&kgre_drm_mm_remove_node, "drm" },
};

int kgr_bsc1173663_gem_resolve_kallsyms(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}

#endif
