/*
 * bsc1173663_gem_execbuffer.c
 *
 * Fix for CVE-2019-0155, bsc#1173663 (i915_gem_execbuffer.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)
#include <linux/dma_remapping.h>
#include <linux/uaccess.h>
#include "bsc1173663_common.h"
#include "kgr_patch_bsc1173663.h"
#include "../kallsyms_relocs.h"

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_batch_pool.h */
static struct drm_i915_gem_object*
(*kgre_i915_gem_batch_pool_get)(struct i915_gem_batch_pool *pool, size_t size);

/* klp-ccp: from drivers/gpu/drm/i915/intel_ringbuffer.h */
static bool (*kgre_intel_ring_initialized)(struct intel_engine_cs *ring);

/* klp-ccp: from drivers/gpu/drm/i915/intel_lrc.h */
static int (*kgre_intel_lr_context_deferred_alloc)(struct intel_context *ctx,
				    struct intel_engine_cs *ring);

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.h */
static const struct i915_ggtt_view (*kgre_i915_ggtt_view_normal);

/* klp-ccp: from include/drm/drm_gem.h */
static void (*kgre_drm_gem_object_free)(struct kref *kref);

static inline void
kgrr_drm_gem_object_unreference(struct drm_gem_object *obj)
{
	if (obj != NULL) {
		WARN_ON(!mutex_is_locked(&obj->dev->struct_mutex));

		kref_put(&obj->refcount, (*kgre_drm_gem_object_free));
	}
}

/* klp-ccp: from drivers/gpu/drm/i915/i915_drv.h */
#undef WARN_ON
#define WARN_ON(x) WARN((x), "WARN_ON(%s)", #x )

static int (*kgre_i915_gem_request_alloc)(struct intel_engine_cs *ring,
			   struct intel_context *ctx,
			   struct drm_i915_gem_request **req_out);
static void (*kgre_i915_gem_request_cancel)(struct drm_i915_gem_request *req);

static int (*kgre_i915_gem_request_add_to_client)(struct drm_i915_gem_request *req,
				   struct drm_file *file);

static void (*kgre_intel_runtime_pm_get)(struct drm_i915_private *dev_priv);

static void (*kgre_intel_runtime_pm_put)(struct drm_i915_private *dev_priv);

static int __must_check
(*kgre_i915_gem_object_pin)(struct drm_i915_gem_object *obj,
		    struct i915_address_space *vm,
		    uint32_t alignment,
		    uint64_t flags);

static int __must_check (*kgre_i915_vma_unbind)(struct i915_vma *vma);

static inline void i915_gem_object_unpin_pages(struct drm_i915_gem_object *obj)
{
	BUG_ON(obj->pages_pin_count == 0);
	obj->pages_pin_count--;
}

static int __must_check (*kgre_i915_mutex_lock_interruptible)(struct drm_device *dev);

static void (*kgre_i915_gem_retire_requests_ring)(struct intel_engine_cs *ring);

static u64 (*kgre_i915_gem_obj_ggtt_offset_view)(struct drm_i915_gem_object *o,
				  const struct i915_ggtt_view *view);
static u64 (*kgre_i915_gem_obj_offset)(struct drm_i915_gem_object *o,
			struct i915_address_space *vm);
static inline u64
kgrr_i915_gem_obj_ggtt_offset(struct drm_i915_gem_object *o)
{
	return (*kgre_i915_gem_obj_ggtt_offset_view)(o, &(*kgre_i915_ggtt_view_normal));
}

static struct i915_vma *
(*kgre_i915_gem_obj_to_vma)(struct drm_i915_gem_object *obj,
		    struct i915_address_space *vm);

static struct i915_vma *
(*kgre_i915_gem_obj_lookup_or_create_vma)(struct drm_i915_gem_object *obj,
				  struct i915_address_space *vm);


static inline int __must_check
kgrr_i915_gem_obj_ggtt_pin(struct drm_i915_gem_object *obj,
		      uint32_t alignment,
		      unsigned flags)
{
	return (*kgre_i915_gem_object_pin)(obj, i915_obj_to_ggtt(obj),
				   alignment, flags | PIN_GLOBAL);
}

static void (*kgre_i915_gem_object_ggtt_unpin_view)(struct drm_i915_gem_object *obj,
				     const struct i915_ggtt_view *view);
static inline void
kgrr_i915_gem_object_ggtt_unpin(struct drm_i915_gem_object *obj)
{
	(*kgre_i915_gem_object_ggtt_unpin_view)(obj, &(*kgre_i915_ggtt_view_normal));
}

static int __must_check (*kgre_i915_gem_object_get_fence)(struct drm_i915_gem_object *obj);

static bool (*kgre_i915_gem_object_pin_fence)(struct drm_i915_gem_object *obj);

static struct intel_context *
(*kgre_i915_gem_context_get)(struct drm_i915_file_private *file_priv, u32 id);
static void (*kgre_i915_gem_context_free)(struct kref *ctx_ref);

static inline void i915_gem_context_reference(struct intel_context *ctx)
{
	kref_get(&ctx->ref);
}

static inline void kgrr_i915_gem_context_unreference(struct intel_context *ctx)
{
	kref_put(&ctx->ref, (*kgre_i915_gem_context_free));
}

static int (*kgre_i915_gem_evict_vm)(struct i915_address_space *vm, bool do_idle);

static bool (*kgre_i915_needs_cmd_parser)(struct intel_engine_cs *ring);

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_execbuffer.c */
#define  __EXEC_OBJECT_HAS_PIN (1<<31)
#define  __EXEC_OBJECT_HAS_FENCE (1<<30)
#define  __EXEC_OBJECT_NEEDS_MAP (1<<29)
#define  __EXEC_OBJECT_NEEDS_BIAS (1<<28)

#define BATCH_OFFSET_BIAS (256*1024)

struct eb_vmas {
	struct list_head vmas;
	int and;
	union {
		struct i915_vma *lut[0];
		struct hlist_head buckets[0];
	};
};

static struct eb_vmas *
eb_create(struct drm_i915_gem_execbuffer2 *args)
{
	struct eb_vmas *eb = NULL;

	if (args->flags & I915_EXEC_HANDLE_LUT) {
		unsigned size = args->buffer_count;
		size *= sizeof(struct i915_vma *);
		size += sizeof(struct eb_vmas);
		eb = kmalloc(size, GFP_TEMPORARY | __GFP_NOWARN | __GFP_NORETRY);
	}

	if (eb == NULL) {
		unsigned size = args->buffer_count;
		unsigned count = PAGE_SIZE / sizeof(struct hlist_head) / 2;
		BUILD_BUG_ON_NOT_POWER_OF_2(PAGE_SIZE / sizeof(struct hlist_head));
		while (count > 2*size)
			count >>= 1;
		eb = kzalloc(count*sizeof(struct hlist_head) +
			     sizeof(struct eb_vmas),
			     GFP_TEMPORARY);
		if (eb == NULL)
			return eb;

		eb->and = count - 1;
	} else
		eb->and = -args->buffer_count;

	INIT_LIST_HEAD(&eb->vmas);
	return eb;
}

static void
eb_reset(struct eb_vmas *eb)
{
	if (eb->and >= 0)
		memset(eb->buckets, 0, (eb->and+1)*sizeof(struct hlist_head));
}

static int
kgrr_eb_lookup_vmas(struct eb_vmas *eb,
	       struct drm_i915_gem_exec_object2 *exec,
	       const struct drm_i915_gem_execbuffer2 *args,
	       struct i915_address_space *vm,
	       struct drm_file *file)
{
	struct drm_i915_gem_object *obj;
	struct list_head objects;
	int i, ret;

	INIT_LIST_HEAD(&objects);
	spin_lock(&file->table_lock);
	/* Grab a reference to the object and release the lock so we can lookup
	 * or create the VMA without using GFP_ATOMIC */
	for (i = 0; i < args->buffer_count; i++) {
		obj = to_intel_bo(idr_find(&file->object_idr, exec[i].handle));
		if (obj == NULL) {
			spin_unlock(&file->table_lock);
			KGRR_DRM_DEBUG("Invalid object handle %d at index %d\n",
				   exec[i].handle, i);
			ret = -ENOENT;
			goto err;
		}

		if (!list_empty(&obj->obj_exec_link)) {
			spin_unlock(&file->table_lock);
			KGRR_DRM_DEBUG("Object %p [handle %d, index %d] appears more than once in object list\n",
				   obj, exec[i].handle, i);
			ret = -EINVAL;
			goto err;
		}

		drm_gem_object_reference(&obj->base);
		list_add_tail(&obj->obj_exec_link, &objects);
	}
	spin_unlock(&file->table_lock);

	i = 0;
	while (!list_empty(&objects)) {
		struct i915_vma *vma;

		obj = list_first_entry(&objects,
				       struct drm_i915_gem_object,
				       obj_exec_link);

		/*
		 * NOTE: We can leak any vmas created here when something fails
		 * later on. But that's no issue since vma_unbind can deal with
		 * vmas which are not actually bound. And since only
		 * lookup_or_create exists as an interface to get at the vma
		 * from the (obj, vm) we don't run the risk of creating
		 * duplicated vmas for the same vm.
		 */
		vma = (*kgre_i915_gem_obj_lookup_or_create_vma)(obj, vm);
		if (IS_ERR(vma)) {
			KGRR_DRM_DEBUG("Failed to lookup VMA\n");
			ret = PTR_ERR(vma);
			goto err;
		}

		/* Transfer ownership from the objects list to the vmas list. */
		list_add_tail(&vma->exec_list, &eb->vmas);
		list_del_init(&obj->obj_exec_link);

		vma->exec_entry = &exec[i];
		if (eb->and < 0) {
			eb->lut[i] = vma;
		} else {
			uint32_t handle = args->flags & I915_EXEC_HANDLE_LUT ? i : exec[i].handle;
			vma->exec_handle = handle;
			hlist_add_head(&vma->exec_node,
				       &eb->buckets[handle & eb->and]);
		}
		++i;
	}

	return 0;


err:
	while (!list_empty(&objects)) {
		obj = list_first_entry(&objects,
				       struct drm_i915_gem_object,
				       obj_exec_link);
		list_del_init(&obj->obj_exec_link);
		kgrr_drm_gem_object_unreference(&obj->base);
	}
	/*
	 * Objects already transfered to the vmas list will be unreferenced by
	 * eb_destroy.
	 */

	return ret;
}

static void
(*kgre_i915_gem_execbuffer_unreserve_vma)(struct i915_vma *vma);

static void kgrr_eb_destroy(struct eb_vmas *eb)
{
	while (!list_empty(&eb->vmas)) {
		struct i915_vma *vma;

		vma = list_first_entry(&eb->vmas,
				       struct i915_vma,
				       exec_list);
		list_del_init(&vma->exec_list);
		(*kgre_i915_gem_execbuffer_unreserve_vma)(vma);
		kgrr_drm_gem_object_unreference(&vma->obj->base);
	}
	kfree(eb);
}

static int
(*kgre_i915_gem_execbuffer_relocate_entry)(struct drm_i915_gem_object *obj,
				   struct eb_vmas *eb,
				   struct drm_i915_gem_relocation_entry *reloc);

static int
kgrr_i915_gem_execbuffer_relocate_vma(struct i915_vma *vma,
				 struct eb_vmas *eb)
{
#define N_RELOC(x) ((x) / sizeof(struct drm_i915_gem_relocation_entry))
	struct drm_i915_gem_relocation_entry stack_reloc[N_RELOC(512)];
	struct drm_i915_gem_relocation_entry __user *user_relocs;
	struct drm_i915_gem_exec_object2 *entry = vma->exec_entry;
	int remain, ret;

	user_relocs = to_user_ptr(entry->relocs_ptr);

	remain = entry->relocation_count;
	while (remain) {
		struct drm_i915_gem_relocation_entry *r = stack_reloc;
		int count = remain;
		if (count > ARRAY_SIZE(stack_reloc))
			count = ARRAY_SIZE(stack_reloc);
		remain -= count;

		if (__copy_from_user_inatomic(r, user_relocs, count*sizeof(r[0])))
			return -EFAULT;

		do {
			u64 offset = r->presumed_offset;

			ret = (*kgre_i915_gem_execbuffer_relocate_entry)(vma->obj, eb, r);
			if (ret)
				return ret;

			if (r->presumed_offset != offset &&
			    __copy_to_user_inatomic(&user_relocs->presumed_offset,
						    &r->presumed_offset,
						    sizeof(r->presumed_offset))) {
				return -EFAULT;
			}

			user_relocs++;
			r++;
		} while (--count);
	}

	return 0;
}

static int
kgrr_i915_gem_execbuffer_relocate_vma_slow(struct i915_vma *vma,
				      struct eb_vmas *eb,
				      struct drm_i915_gem_relocation_entry *relocs)
{
	const struct drm_i915_gem_exec_object2 *entry = vma->exec_entry;
	int i, ret;

	for (i = 0; i < entry->relocation_count; i++) {
		ret = (*kgre_i915_gem_execbuffer_relocate_entry)(vma->obj, eb, &relocs[i]);
		if (ret)
			return ret;
	}

	return 0;
}

static int
kgrr_i915_gem_execbuffer_relocate(struct eb_vmas *eb)
{
	struct i915_vma *vma;
	int ret = 0;

	/* This is the fast path and we cannot handle a pagefault whilst
	 * holding the struct mutex lest the user pass in the relocations
	 * contained within a mmaped bo. For in such a case we, the page
	 * fault handler would call i915_gem_fault() and we would try to
	 * acquire the struct mutex again. Obviously this is bad and so
	 * lockdep complains vehemently.
	 */
	pagefault_disable();
	list_for_each_entry(vma, &eb->vmas, exec_list) {
		ret = kgrr_i915_gem_execbuffer_relocate_vma(vma, eb);
		if (ret)
			break;
	}
	pagefault_enable();

	return ret;
}

static bool only_mappable_for_reloc(unsigned int flags)
{
	return (flags & (EXEC_OBJECT_NEEDS_FENCE | __EXEC_OBJECT_NEEDS_MAP)) ==
		__EXEC_OBJECT_NEEDS_MAP;
}

static int
kgrr_i915_gem_execbuffer_reserve_vma(struct i915_vma *vma,
				struct intel_engine_cs *ring,
				bool *need_reloc)
{
	struct drm_i915_gem_object *obj = vma->obj;
	struct drm_i915_gem_exec_object2 *entry = vma->exec_entry;
	uint64_t flags;
	int ret;

	flags = PIN_USER;
	if (entry->flags & EXEC_OBJECT_NEEDS_GTT)
		flags |= PIN_GLOBAL;

	if (!drm_mm_node_allocated(&vma->node)) {
		/* Wa32bitGeneralStateOffset & Wa32bitInstructionBaseOffset,
		 * limit address to the first 4GBs for unflagged objects.
		 */
		if ((entry->flags & EXEC_OBJECT_SUPPORTS_48B_ADDRESS) == 0)
			flags |= PIN_ZONE_4G;
		if (entry->flags & __EXEC_OBJECT_NEEDS_MAP)
			flags |= PIN_GLOBAL | PIN_MAPPABLE;
		if (entry->flags & __EXEC_OBJECT_NEEDS_BIAS)
			flags |= BATCH_OFFSET_BIAS | PIN_OFFSET_BIAS;
		if ((flags & PIN_MAPPABLE) == 0)
			flags |= PIN_HIGH;
	}

	ret = (*kgre_i915_gem_object_pin)(obj, vma->vm, entry->alignment, flags);
	if ((ret == -ENOSPC  || ret == -E2BIG) &&
	    only_mappable_for_reloc(entry->flags))
		ret = (*kgre_i915_gem_object_pin)(obj, vma->vm,
					  entry->alignment,
					  flags & ~PIN_MAPPABLE);
	if (ret)
		return ret;

	entry->flags |= __EXEC_OBJECT_HAS_PIN;

	if (entry->flags & EXEC_OBJECT_NEEDS_FENCE) {
		ret = (*kgre_i915_gem_object_get_fence)(obj);
		if (ret)
			return ret;

		if ((*kgre_i915_gem_object_pin_fence)(obj))
			entry->flags |= __EXEC_OBJECT_HAS_FENCE;
	}

	if (entry->offset != vma->node.start) {
		entry->offset = vma->node.start;
		*need_reloc = true;
	}

	if (entry->flags & EXEC_OBJECT_WRITE) {
		obj->base.pending_read_domains = I915_GEM_DOMAIN_RENDER;
		obj->base.pending_write_domain = I915_GEM_DOMAIN_RENDER;
	}

	return 0;
}

static bool
need_reloc_mappable(struct i915_vma *vma)
{
	struct drm_i915_gem_exec_object2 *entry = vma->exec_entry;

	if (entry->relocation_count == 0)
		return false;

	if (!i915_is_ggtt(vma->vm))
		return false;

	/* See also use_cpu_reloc() */
	if (HAS_LLC(vma->obj->base.dev))
		return false;

	if (vma->obj->base.write_domain == I915_GEM_DOMAIN_CPU)
		return false;

	return true;
}

static bool
eb_vma_misplaced(struct i915_vma *vma)
{
	struct drm_i915_gem_exec_object2 *entry = vma->exec_entry;
	struct drm_i915_gem_object *obj = vma->obj;

	WARN_ON(entry->flags & __EXEC_OBJECT_NEEDS_MAP &&
	       !i915_is_ggtt(vma->vm));

	if (entry->alignment &&
	    vma->node.start & (entry->alignment - 1))
		return true;

	if (entry->flags & __EXEC_OBJECT_NEEDS_BIAS &&
	    vma->node.start < BATCH_OFFSET_BIAS)
		return true;

	/* avoid costly ping-pong once a batch bo ended up non-mappable */
	if (entry->flags & __EXEC_OBJECT_NEEDS_MAP && !obj->map_and_fenceable)
		return !only_mappable_for_reloc(entry->flags);

	if ((entry->flags & EXEC_OBJECT_SUPPORTS_48B_ADDRESS) == 0 &&
	    (vma->node.start + vma->node.size - 1) >> 32)
		return true;

	return false;
}

static int
kgrr_i915_gem_execbuffer_reserve(struct intel_engine_cs *ring,
			    struct list_head *vmas,
			    struct intel_context *ctx,
			    bool *need_relocs)
{
	struct drm_i915_gem_object *obj;
	struct i915_vma *vma;
	struct i915_address_space *vm;
	struct list_head ordered_vmas;
	bool has_fenced_gpu_access = INTEL_INFO(ring->dev)->gen < 4;
	int retry;

	(*kgre_i915_gem_retire_requests_ring)(ring);

	vm = list_first_entry(vmas, struct i915_vma, exec_list)->vm;

	INIT_LIST_HEAD(&ordered_vmas);
	while (!list_empty(vmas)) {
		struct drm_i915_gem_exec_object2 *entry;
		bool need_fence, need_mappable;

		vma = list_first_entry(vmas, struct i915_vma, exec_list);
		obj = vma->obj;
		entry = vma->exec_entry;

		if (ctx->flags & CONTEXT_NO_ZEROMAP)
			entry->flags |= __EXEC_OBJECT_NEEDS_BIAS;

		if (!has_fenced_gpu_access)
			entry->flags &= ~EXEC_OBJECT_NEEDS_FENCE;
		need_fence =
			entry->flags & EXEC_OBJECT_NEEDS_FENCE &&
			obj->tiling_mode != I915_TILING_NONE;
		need_mappable = need_fence || need_reloc_mappable(vma);

		if (need_mappable) {
			entry->flags |= __EXEC_OBJECT_NEEDS_MAP;
			list_move(&vma->exec_list, &ordered_vmas);
		} else
			list_move_tail(&vma->exec_list, &ordered_vmas);

		obj->base.pending_read_domains = I915_GEM_GPU_DOMAINS & ~I915_GEM_DOMAIN_COMMAND;
		obj->base.pending_write_domain = 0;
	}
	list_splice(&ordered_vmas, vmas);

	/* Attempt to pin all of the buffers into the GTT.
	 * This is done in 3 phases:
	 *
	 * 1a. Unbind all objects that do not match the GTT constraints for
	 *     the execbuffer (fenceable, mappable, alignment etc).
	 * 1b. Increment pin count for already bound objects.
	 * 2.  Bind new objects.
	 * 3.  Decrement pin count.
	 *
	 * This avoid unnecessary unbinding of later objects in order to make
	 * room for the earlier objects *unless* we need to defragment.
	 */
	retry = 0;
	do {
		int ret = 0;

		/* Unbind any ill-fitting objects or pin. */
		list_for_each_entry(vma, vmas, exec_list) {
			if (!drm_mm_node_allocated(&vma->node))
				continue;

			if (eb_vma_misplaced(vma))
				ret = (*kgre_i915_vma_unbind)(vma);
			else
				ret = kgrr_i915_gem_execbuffer_reserve_vma(vma, ring, need_relocs);
			if (ret)
				goto err;
		}

		/* Bind fresh objects */
		list_for_each_entry(vma, vmas, exec_list) {
			if (drm_mm_node_allocated(&vma->node))
				continue;

			ret = kgrr_i915_gem_execbuffer_reserve_vma(vma, ring, need_relocs);
			if (ret)
				goto err;
		}

err:
		if (ret != -ENOSPC || retry++)
			return ret;

		/* Decrement pin count for bound objects */
		list_for_each_entry(vma, vmas, exec_list)
			(*kgre_i915_gem_execbuffer_unreserve_vma)(vma);

		ret = (*kgre_i915_gem_evict_vm)(vm, true);
		if (ret)
			return ret;
	} while (1);
}

static int
kgrr_i915_gem_execbuffer_relocate_slow(struct drm_device *dev,
				  struct drm_i915_gem_execbuffer2 *args,
				  struct drm_file *file,
				  struct intel_engine_cs *ring,
				  struct eb_vmas *eb,
				  struct drm_i915_gem_exec_object2 *exec,
				  struct intel_context *ctx)
{
	struct drm_i915_gem_relocation_entry *reloc;
	struct i915_address_space *vm;
	struct i915_vma *vma;
	bool need_relocs;
	int *reloc_offset;
	int i, total, ret;
	unsigned count = args->buffer_count;

	vm = list_first_entry(&eb->vmas, struct i915_vma, exec_list)->vm;

	/* We may process another execbuffer during the unlock... */
	while (!list_empty(&eb->vmas)) {
		vma = list_first_entry(&eb->vmas, struct i915_vma, exec_list);
		list_del_init(&vma->exec_list);
		(*kgre_i915_gem_execbuffer_unreserve_vma)(vma);
		kgrr_drm_gem_object_unreference(&vma->obj->base);
	}

	mutex_unlock(&dev->struct_mutex);

	total = 0;
	for (i = 0; i < count; i++)
		total += exec[i].relocation_count;

	reloc_offset = drm_malloc_ab(count, sizeof(*reloc_offset));
	reloc = drm_malloc_ab(total, sizeof(*reloc));
	if (reloc == NULL || reloc_offset == NULL) {
		drm_free_large(reloc);
		drm_free_large(reloc_offset);
		mutex_lock(&dev->struct_mutex);
		return -ENOMEM;
	}

	total = 0;
	for (i = 0; i < count; i++) {
		struct drm_i915_gem_relocation_entry __user *user_relocs;
		u64 invalid_offset = (u64)-1;
		int j;

		user_relocs = to_user_ptr(exec[i].relocs_ptr);

		if (copy_from_user(reloc+total, user_relocs,
				   exec[i].relocation_count * sizeof(*reloc))) {
			ret = -EFAULT;
			mutex_lock(&dev->struct_mutex);
			goto err;
		}

		/* As we do not update the known relocation offsets after
		 * relocating (due to the complexities in lock handling),
		 * we need to mark them as invalid now so that we force the
		 * relocation processing next time. Just in case the target
		 * object is evicted and then rebound into its old
		 * presumed_offset before the next execbuffer - if that
		 * happened we would make the mistake of assuming that the
		 * relocations were valid.
		 */
		for (j = 0; j < exec[i].relocation_count; j++) {
			if (__copy_to_user(&user_relocs[j].presumed_offset,
					   &invalid_offset,
					   sizeof(invalid_offset))) {
				ret = -EFAULT;
				mutex_lock(&dev->struct_mutex);
				goto err;
			}
		}

		reloc_offset[i] = total;
		total += exec[i].relocation_count;
	}

	ret = (*kgre_i915_mutex_lock_interruptible)(dev);
	if (ret) {
		mutex_lock(&dev->struct_mutex);
		goto err;
	}

	/* reacquire the objects */
	eb_reset(eb);
	ret = kgrr_eb_lookup_vmas(eb, exec, args, vm, file);
	if (ret)
		goto err;

	need_relocs = (args->flags & I915_EXEC_NO_RELOC) == 0;
	ret = kgrr_i915_gem_execbuffer_reserve(ring, &eb->vmas, ctx, &need_relocs);
	if (ret)
		goto err;

	list_for_each_entry(vma, &eb->vmas, exec_list) {
		int offset = vma->exec_entry - exec;
		ret = kgrr_i915_gem_execbuffer_relocate_vma_slow(vma, eb,
							    reloc + reloc_offset[offset]);
		if (ret)
			goto err;
	}

	/* Leave the user relocations as are, this is the painfully slow path,
	 * and we want to avoid the complication of dropping the lock whilst
	 * having buffers reserved in the aperture and so causing spurious
	 * ENOSPC for random operations.
	 */

err:
	drm_free_large(reloc);
	drm_free_large(reloc_offset);
	return ret;
}

static bool
kgrr_i915_gem_check_execbuffer(struct drm_i915_gem_execbuffer2 *exec)
{
	if (exec->flags & __I915_EXEC_UNKNOWN_FLAGS)
		return false;

	/* Kernel clipping was a DRI1 misfeature */
	if (exec->num_cliprects || exec->cliprects_ptr)
		return false;

	if (exec->DR4 == 0xffffffff) {
		KGRR_DRM_DEBUG("UXA submitting garbage DR4, fixing up\n");
		exec->DR4 = 0;
	}
	if (exec->DR1 || exec->DR4)
		return false;

	if ((exec->batch_start_offset | exec->batch_len) & 0x7)
		return false;

	return true;
}

static int
kgrr_validate_exec_list(struct drm_device *dev,
		   struct drm_i915_gem_exec_object2 *exec,
		   int count)
{
	unsigned relocs_total = 0;
	unsigned relocs_max = UINT_MAX / sizeof(struct drm_i915_gem_relocation_entry);
	unsigned invalid_flags;
	int i;

	invalid_flags = __EXEC_OBJECT_UNKNOWN_FLAGS;
	if (((*kgre_i915).enable_ppgtt >= 2))
		invalid_flags |= EXEC_OBJECT_NEEDS_GTT;

	for (i = 0; i < count; i++) {
		char __user *ptr = to_user_ptr(exec[i].relocs_ptr);
		int length; /* limited by fault_in_pages_readable() */

		if (exec[i].flags & invalid_flags)
			return -EINVAL;

		if (exec[i].alignment && !is_power_of_2(exec[i].alignment))
			return -EINVAL;

		/* First check for malicious input causing overflow in
		 * the worst case where we need to allocate the entire
		 * relocation tree as a single array.
		 */
		if (exec[i].relocation_count > relocs_max - relocs_total)
			return -EINVAL;
		relocs_total += exec[i].relocation_count;

		length = exec[i].relocation_count *
			sizeof(struct drm_i915_gem_relocation_entry);
		/*
		 * We must check that the entire relocation array is safe
		 * to read, but since we may need to update the presumed
		 * offsets during execution, check for full write access.
		 */
		if (!access_ok(VERIFY_WRITE, ptr, length))
			return -EFAULT;

		if (likely(!(*kgre_i915).prefault_disable)) {
			if (fault_in_multipages_readable(ptr, length))
				return -EFAULT;
		}
	}

	return 0;
}

static struct intel_context *
kgrr_i915_gem_validate_context(struct drm_device *dev, struct drm_file *file,
			  struct intel_engine_cs *ring, const u32 ctx_id)
{
	struct intel_context *ctx = NULL;
	struct i915_ctx_hang_stats *hs;

	if (ring->id != RCS && ctx_id != DEFAULT_CONTEXT_HANDLE)
		return ERR_PTR(-EINVAL);

	ctx = (*kgre_i915_gem_context_get)(file->driver_priv, ctx_id);
	if (IS_ERR(ctx))
		return ctx;

	hs = &ctx->hang_stats;
	if (hs->banned) {
		KGRR_DRM_DEBUG("Context %u tried to submit while banned\n", ctx_id);
		return ERR_PTR(-EIO);
	}

	if ((*kgre_i915).enable_execlists && !ctx->engine[ring->id].state) {
		int ret = (*kgre_intel_lr_context_deferred_alloc)(ctx, ring);
		if (ret) {
			KGRR_DRM_DEBUG("Could not create LRC %u: %d\n", ctx_id, ret);
			return ERR_PTR(ret);
		}
	}

	return ctx;
}

/*
 * New.
 * Notes:
 * - We reach here only for gen7 and gen9 devices, everything else
 *   doesn't invoke the cmd parser.
 * - The only caller, i915_gem_do_execbuffer() passes either
 *   ctx->ppgtt as vm if !NULL or the device's ggtt.
 * - ctx->ppgtt is set iff USES_FULL_PPGTT(dev) returns true,
 *   c.f. i915_gem_create_context().
 * - USES_FULL_PPGTT(dev) is set if enable_ppgtt >= 2.
 * - enable_ppgtt is defaulted in sanitize_enable_ppgtt() to 2 on
 *   gen8+ and to 1 or zero before, but can be overriden by users.
 *
 * Upstream checks for the new vm->has_read_only, but in order to
 * avoid shadow variables, we attempt to emulate that by an equivalent
 * condition as far as the context here is concerned. ->has_read_only
 * gets set for IS_VALLEYVIEW() GGTTs and gen8+ ppgts (latter only if
 * !intel_vgpu_active()). IS_VALLEYVIEW() devices of gen7 and gen8
 * exist, but gen7 is caught by the CMDPARSER_USES_GGTT() if clause
 * and for gen8 the parser isn't executed. In conclusion, the vm
 * instance in the else-if clause below is a ppgtt if enable_ppgtt >=
 * 2 and it's always gen8+ (gen9 actually). It's ->has_read_only would
 * be set if !intel_vgpu_active(). With the later upstream commit
 * e5ee4956f2fd ("drm/i915/gtt: Revert "Disable read-only support
 * under GVT"), even the condition on intel_vgpu_active() had been
 * removed.
 */
static struct i915_vma*
kgrp_shadow_batch_pin(struct drm_i915_gem_object *obj, struct i915_address_space *vm)
{
	struct drm_i915_private *dev_priv = to_i915(obj->base.dev);
	struct i915_address_space *pin_vm = vm;
	u64 flags;
	int ret;

	/*
	 * PPGTT backed shadow buffers must be mapped RO, to prevent
	 * post-scan tampering
	 */
	if (KGRP_CMDPARSER_USES_GGTT(dev_priv)) {
		flags = PIN_GLOBAL;
		pin_vm = &dev_priv->gtt.base;
		ret = kgrr_i915_gem_obj_ggtt_pin(obj, 0, 0);
	} else if ((*kgre_i915).enable_ppgtt >= 2) { /* vm->has_read_only */
		flags = PIN_USER;
		obj->gt_ro = 1;
		/*
		 * Call the patched variant capable of clearing
		 * _PAGE_RW on gen8, i.e. which invokes the patched
		 * gen8_pte_encode() down the callchain.
		 */
		ret = kgrp_i915_gem_object_pin(obj, pin_vm, 0, flags);
	} else {
		KGRR_DRM_DEBUG("Cannot prevent post-scan tampering without RO capable vm\n");
		return ERR_PTR(-EINVAL);
	}

	if (ret)
		return ERR_PTR(ret);
	else
		return (*kgre_i915_gem_obj_to_vma)(obj, pin_vm);
}

static struct drm_i915_gem_object*
/*
 * Fix CVE-2019-0155
 *  -1 line, +2 lines
 */
kgrp_i915_gem_execbuffer_parse(struct intel_context *ctx,
			  struct intel_engine_cs *ring,
			  struct drm_i915_gem_exec_object2 *shadow_exec_entry,
			  struct eb_vmas *eb,
/*
 * Fix CVE-2019-0155
 *  +1 line
 */
			  struct i915_address_space *vm,
			  struct drm_i915_gem_object *batch_obj,
			  u32 batch_start_offset,
/*
 * Fix CVE-2019-0155
 *  -2 lines, +1 line
 */
			  u32 batch_len)
{
	struct drm_i915_gem_object *shadow_batch_obj;
	struct i915_vma *vma;
	/*
	 * Fix CVE-2019-0155
	 *  +4 lines
	 */
	struct i915_vma *user_vma = list_entry(eb->vmas.prev,
					typeof(*user_vma), exec_list);
	u64 batch_start;
	u64 shadow_batch_start;
	int ret;

	shadow_batch_obj = (*kgre_i915_gem_batch_pool_get)(&ring->batch_pool,
						   PAGE_ALIGN(batch_len));
	if (IS_ERR(shadow_batch_obj))
		return shadow_batch_obj;

	/*
	 * Fix CVE-2019-0155
	 *  +5 lines
	 */
	vma = kgrp_shadow_batch_pin(shadow_batch_obj, vm);
	if (IS_ERR(vma)) {
		ret = PTR_ERR(vma);
		goto err;
	}

	/*
	 * Fix CVE-2019-0155
	 *  +2 lines
	 */
	batch_start = user_vma->node.start + batch_start_offset;
	shadow_batch_start = vma->node.start;

	/*
	 * Fix CVE-2019-0155
	 *  -8 lines, +13 lines
	 */
	ret = kgrp_i915_parse_cmds(ctx,
			      ring,
			      batch_obj,
			      batch_start,
			      batch_start_offset,
			      batch_len,
			      shadow_batch_obj,
			      shadow_batch_start);
	if (ret) {
		WARN_ON(vma->pin_count == 0);
		vma->pin_count--;
		goto err;
	}

	/*
	 * Fix CVE-2019-0155
	 *  -3 lines
	 */

	i915_gem_object_unpin_pages(shadow_batch_obj);

	memset(shadow_exec_entry, 0, sizeof(*shadow_exec_entry));

	/*
	 * Fix CVE-2019-0155
	 *  -1 line
	 */
	vma->exec_entry = shadow_exec_entry;
	vma->exec_entry->flags = __EXEC_OBJECT_HAS_PIN;
	drm_gem_object_reference(&shadow_batch_obj->base);
	list_add_tail(&vma->exec_list, &eb->vmas);

	shadow_batch_obj->base.pending_read_domains = I915_GEM_DOMAIN_COMMAND;

	return shadow_batch_obj;

err:
	i915_gem_object_unpin_pages(shadow_batch_obj);
	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +1 line
	 */
	/*
	 * Unsafe GGTT-backed buffers can still be submitted safely
	 * as non-secure.
	 * For PPGTT backing however, we have no choice but to forcibly
	 * reject unsafe buffers
	 */
	if (KGRP_CMDPARSER_USES_GGTT(batch_obj->base.dev) && (ret == -EACCES))
		return batch_obj;
	else
		return ERR_PTR(ret);
}

static int gen8_dispatch_bsd_ring(struct drm_device *dev,
				  struct drm_file *file)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	struct drm_i915_file_private *file_priv = file->driver_priv;

	/* Check whether the file_priv is using one ring */
	if (file_priv->bsd_ring)
		return file_priv->bsd_ring->id;
	else {
		/* If no, use the ping-pong mechanism to select one ring */
		int ring_id;

		mutex_lock(&dev->struct_mutex);
		if (dev_priv->mm.bsd_ring_dispatch_index == 0) {
			ring_id = VCS;
			dev_priv->mm.bsd_ring_dispatch_index = 1;
		} else {
			ring_id = VCS2;
			dev_priv->mm.bsd_ring_dispatch_index = 0;
		}
		file_priv->bsd_ring = &dev_priv->ring[ring_id];
		mutex_unlock(&dev->struct_mutex);
		return ring_id;
	}
}

static struct drm_i915_gem_object *
eb_get_batch(struct eb_vmas *eb)
{
	struct i915_vma *vma = list_entry(eb->vmas.prev, typeof(*vma), exec_list);

	/*
	 * SNA is doing fancy tricks with compressing batch buffers, which leads
	 * to negative relocation deltas. Usually that works out ok since the
	 * relocate address is still positive, except when the batch is placed
	 * very low in the GTT. Ensure this doesn't happen.
	 *
	 * Note that actual hangs have only been observed on gen7, but for
	 * paranoia do it everywhere.
	 */
	vma->exec_entry->flags |= __EXEC_OBJECT_NEEDS_BIAS;

	return vma->obj;
}

static int
kgrp_i915_gem_do_execbuffer(struct drm_device *dev, void *data,
		       struct drm_file *file,
		       struct drm_i915_gem_execbuffer2 *args,
		       struct drm_i915_gem_exec_object2 *exec)
{
	struct drm_i915_private *dev_priv = dev->dev_private;
	struct eb_vmas *eb;
	struct drm_i915_gem_object *batch_obj;
	struct drm_i915_gem_exec_object2 shadow_exec_entry;
	struct intel_engine_cs *ring;
	struct intel_context *ctx;
	struct i915_address_space *vm;
	struct i915_execbuffer_params params_master; /* XXX: will be removed later */
	struct i915_execbuffer_params *params = &params_master;
	const u32 ctx_id = i915_execbuffer2_get_context_id(*args);
	u32 dispatch_flags;
	int ret;
	bool need_relocs;

	if (!kgrr_i915_gem_check_execbuffer(args))
		return -EINVAL;

	ret = kgrr_validate_exec_list(dev, exec, args->buffer_count);
	if (ret)
		return ret;

	dispatch_flags = 0;
	if (args->flags & I915_EXEC_SECURE) {
		/*
		 * Fix CVE-2019-0155
		 *  +-0 lines
		 * Upstream returns -EPERM here for gen6+.
		 * But we don't want to break existing applications
		 * from a livepatch and continue to let root shoot
		 * him/hersef into the foot, if he/she wants to.
		 * In order for this to continue to work, we must
		 * invoke the parser for gen9 below only if
		 * I915_EXEC_SECURE had not been set -- otherwise
		 * batch_obj will be made to point to
		 * shadow_batch_obj.
		 */
		if (!file->is_master || !capable(CAP_SYS_ADMIN))
		    return -EPERM;

		dispatch_flags |= I915_DISPATCH_SECURE;
	}
	if (args->flags & I915_EXEC_IS_PINNED)
		dispatch_flags |= I915_DISPATCH_PINNED;

	if ((args->flags & I915_EXEC_RING_MASK) > LAST_USER_RING) {
		KGRR_DRM_DEBUG("execbuf with unknown ring: %d\n",
			  (int)(args->flags & I915_EXEC_RING_MASK));
		return -EINVAL;
	}

	if (((args->flags & I915_EXEC_RING_MASK) != I915_EXEC_BSD) &&
	    ((args->flags & I915_EXEC_BSD_MASK) != 0)) {
		KGRR_DRM_DEBUG("execbuf with non bsd ring but with invalid "
			"bsd dispatch flags: %d\n", (int)(args->flags));
		return -EINVAL;
	} 

	if ((args->flags & I915_EXEC_RING_MASK) == I915_EXEC_DEFAULT)
		ring = &dev_priv->ring[RCS];
	else if ((args->flags & I915_EXEC_RING_MASK) == I915_EXEC_BSD) {
		if (HAS_BSD2(dev)) {
			int ring_id;

			switch (args->flags & I915_EXEC_BSD_MASK) {
			case I915_EXEC_BSD_DEFAULT:
				ring_id = gen8_dispatch_bsd_ring(dev, file);
				ring = &dev_priv->ring[ring_id];
				break;
			case I915_EXEC_BSD_RING1:
				ring = &dev_priv->ring[VCS];
				break;
			case I915_EXEC_BSD_RING2:
				ring = &dev_priv->ring[VCS2];
				break;
			default:
				KGRR_DRM_DEBUG("execbuf with unknown bsd ring: %d\n",
					  (int)(args->flags & I915_EXEC_BSD_MASK));
				return -EINVAL;
			}
		} else
			ring = &dev_priv->ring[VCS];
	} else
		ring = &dev_priv->ring[(args->flags & I915_EXEC_RING_MASK) - 1];

	if (!(*kgre_intel_ring_initialized)(ring)) {
		KGRR_DRM_DEBUG("execbuf with invalid ring: %d\n",
			  (int)(args->flags & I915_EXEC_RING_MASK));
		return -EINVAL;
	}

	if (args->buffer_count < 1) {
		KGRR_DRM_DEBUG("execbuf with %d buffers\n", args->buffer_count);
		return -EINVAL;
	}

	if (args->flags & I915_EXEC_RESOURCE_STREAMER) {
		if (!HAS_RESOURCE_STREAMER(dev)) {
			KGRR_DRM_DEBUG("RS is only allowed for Haswell, Gen8 and above\n");
			return -EINVAL;
		}
		if (ring->id != RCS) {
			KGRR_DRM_DEBUG("RS is not available on %s\n",
				 ring->name);
			return -EINVAL;
		}

		dispatch_flags |= I915_DISPATCH_RS;
	}

	(*kgre_intel_runtime_pm_get)(dev_priv);

	ret = (*kgre_i915_mutex_lock_interruptible)(dev);
	if (ret)
		goto pre_mutex_err;

	ctx = kgrr_i915_gem_validate_context(dev, file, ring, ctx_id);
	if (IS_ERR(ctx)) {
		mutex_unlock(&dev->struct_mutex);
		ret = PTR_ERR(ctx);
		goto pre_mutex_err;
	}

	i915_gem_context_reference(ctx);

	if (ctx->ppgtt)
		vm = &ctx->ppgtt->base;
	else
		vm = &dev_priv->gtt.base;

	memset(&params_master, 0x00, sizeof(params_master));

	eb = eb_create(args);
	if (eb == NULL) {
		kgrr_i915_gem_context_unreference(ctx);
		mutex_unlock(&dev->struct_mutex);
		ret = -ENOMEM;
		goto pre_mutex_err;
	}

	/* Look up object handles */
	ret = kgrr_eb_lookup_vmas(eb, exec, args, vm, file);
	if (ret)
		goto err;

	/* take note of the batch buffer before we might reorder the lists */
	batch_obj = eb_get_batch(eb);

	/* Move the objects en-masse into the GTT, evicting if necessary. */
	need_relocs = (args->flags & I915_EXEC_NO_RELOC) == 0;
	ret = kgrr_i915_gem_execbuffer_reserve(ring, &eb->vmas, ctx, &need_relocs);
	if (ret)
		goto err;

	/* The objects are in their final locations, apply the relocations. */
	if (need_relocs)
		ret = kgrr_i915_gem_execbuffer_relocate(eb);
	if (ret) {
		if (ret == -EFAULT) {
			ret = kgrr_i915_gem_execbuffer_relocate_slow(dev, args, file, ring,
								eb, exec, ctx);
			BUG_ON(!mutex_is_locked(&dev->struct_mutex));
		}
		if (ret)
			goto err;
	}

	/* Set the pending read domains for the batch buffer to COMMAND */
	if (batch_obj->base.pending_write_domain) {
		KGRR_DRM_DEBUG("Attempting to use self-modifying batch buffer\n");
		ret = -EINVAL;
		goto err;
	}

	params->args_batch_start_offset = args->batch_start_offset;
	/*
	 * Fix CVE-2019-0155
	 *  -1 line, +2 lines
	 * See the comment at the I915_EXEC_SECURE handling code
	 * above: don't invoke the parser for gen9 if root called this
	 * with I915_EXEC_SECURE set. This is important to retain
	 * original behaviour and to not change batch_obj to
	 * shadow_batch_obj.
	 */
	if (((*kgre_i915_needs_cmd_parser)(ring) && args->batch_len) ||
	    (!(dispatch_flags & I915_DISPATCH_SECURE) && kgrp_is_gen9_blt(ring))) {
		struct drm_i915_gem_object *parsed_batch_obj;

		/*
		 * Fix CVE-2019-0155
		 *  +4 lines
		 */
		u32 batch_off = args->batch_start_offset;
		u32 batch_len = args->batch_len;
		if (batch_len == 0)
			batch_len = batch_obj->base.size - batch_off;

		/*
		 * Fix CVE-2019-0155
		 *  -1 line, +1 line
		 */
		parsed_batch_obj = kgrp_i915_gem_execbuffer_parse(ctx, ring,
						      &shadow_exec_entry,
						      /*
						       * Fix CVE-2019-0155
						       *  -1 line, +1 line
						       */
						      eb, vm,
						      batch_obj,
						      /*
						       * Fix CVE-2019-0155
						       *  -3 lines, +2 lines
						       */
						      batch_off,
						      batch_len);
		if (IS_ERR(parsed_batch_obj)) {
			ret = PTR_ERR(parsed_batch_obj);
			goto err;
		}

		/*
		 * parsed_batch_obj == batch_obj means batch not fully parsed:
		 * Accept, but don't promote to secure.
		 */

		if (parsed_batch_obj != batch_obj) {
			/*
			 * Fix CVE-2019-0155
			 *  -10 lines, +2 lines
			 */
			if (KGRP_CMDPARSER_USES_GGTT(dev_priv))
				dispatch_flags |= I915_DISPATCH_SECURE;
			params->args_batch_start_offset = 0;
			batch_obj = parsed_batch_obj;
		}
	}

	batch_obj->base.pending_read_domains |= I915_GEM_DOMAIN_COMMAND;

	/* snb/ivb/vlv conflate the "batch in ppgtt" bit with the "non-secure
	 * batch" bit. Hence we need to pin secure batches into the global gtt.
	 * hsw should have this fixed, but bdw mucks it up again. */
	if (dispatch_flags & I915_DISPATCH_SECURE) {
		/*
		 * So on first glance it looks freaky that we pin the batch here
		 * outside of the reservation loop. But:
		 * - The batch is already pinned into the relevant ppgtt, so we
		 *   already have the backing storage fully allocated.
		 * - No other BO uses the global gtt (well contexts, but meh),
		 *   so we don't really have issues with multiple objects not
		 *   fitting due to fragmentation.
		 * So this is actually safe.
		 */
		ret = kgrr_i915_gem_obj_ggtt_pin(batch_obj, 0, 0);
		if (ret)
			goto err;

		params->batch_obj_vm_offset = kgrr_i915_gem_obj_ggtt_offset(batch_obj);
	} else
		params->batch_obj_vm_offset = (*kgre_i915_gem_obj_offset)(batch_obj, vm);

	/* Allocate a request for this batch buffer nice and early. */
	ret = (*kgre_i915_gem_request_alloc)(ring, ctx, &params->request);
	if (ret)
		goto err_batch_unpin;

	ret = (*kgre_i915_gem_request_add_to_client)(params->request, file);
	if (ret)
		goto err_batch_unpin;

	/*
	 * Save assorted stuff away to pass through to *_submission().
	 * NB: This data should be 'persistent' and not local as it will
	 * kept around beyond the duration of the IOCTL once the GPU
	 * scheduler arrives.
	 */
	params->dev                     = dev;
	params->file                    = file;
	params->ring                    = ring;
	params->dispatch_flags          = dispatch_flags;
	params->batch_obj               = batch_obj;
	params->ctx                     = ctx;

	ret = dev_priv->gt.execbuf_submit(params, args, &eb->vmas);

err_batch_unpin:
	/*
	 * FIXME: We crucially rely upon the active tracking for the (ppgtt)
	 * batch vma for correctness. For less ugly and less fragility this
	 * needs to be adjusted to also track the ggtt batch vma properly as
	 * active.
	 */
	if (dispatch_flags & I915_DISPATCH_SECURE)
		kgrr_i915_gem_object_ggtt_unpin(batch_obj);

err:
	/* the request owns the ref now */
	kgrr_i915_gem_context_unreference(ctx);
	kgrr_eb_destroy(eb);

	/*
	 * If the request was created but not successfully submitted then it
	 * must be freed again. If it was submitted then it is being tracked
	 * on the active request list and no clean up is required here.
	 */
	if (ret && params->request)
		(*kgre_i915_gem_request_cancel)(params->request);

	mutex_unlock(&dev->struct_mutex);

pre_mutex_err:
	/* intel_gpu_busy should also get a ref, so it will free when the device
	 * is really idle. */
	(*kgre_intel_runtime_pm_put)(dev_priv);
	return ret;
}

int
kgrp_i915_gem_execbuffer(struct drm_device *dev, void *data,
		    struct drm_file *file)
{
	struct drm_i915_gem_execbuffer *args = data;
	struct drm_i915_gem_execbuffer2 exec2;
	struct drm_i915_gem_exec_object *exec_list = NULL;
	struct drm_i915_gem_exec_object2 *exec2_list = NULL;
	int ret, i;

	if (args->buffer_count < 1) {
		KGRR_DRM_DEBUG("execbuf with %d buffers\n", args->buffer_count);
		return -EINVAL;
	}

	/* Copy in the exec list from userland */
	exec_list = drm_malloc_ab(sizeof(*exec_list), args->buffer_count);
	exec2_list = drm_malloc_ab(sizeof(*exec2_list), args->buffer_count);
	if (exec_list == NULL || exec2_list == NULL) {
		KGRR_DRM_DEBUG("Failed to allocate exec list for %d buffers\n",
			  args->buffer_count);
		drm_free_large(exec_list);
		drm_free_large(exec2_list);
		return -ENOMEM;
	}
	ret = copy_from_user(exec_list,
			     to_user_ptr(args->buffers_ptr),
			     sizeof(*exec_list) * args->buffer_count);
	if (ret != 0) {
		KGRR_DRM_DEBUG("copy %d exec entries failed %d\n",
			  args->buffer_count, ret);
		drm_free_large(exec_list);
		drm_free_large(exec2_list);
		return -EFAULT;
	}

	for (i = 0; i < args->buffer_count; i++) {
		exec2_list[i].handle = exec_list[i].handle;
		exec2_list[i].relocation_count = exec_list[i].relocation_count;
		exec2_list[i].relocs_ptr = exec_list[i].relocs_ptr;
		exec2_list[i].alignment = exec_list[i].alignment;
		exec2_list[i].offset = exec_list[i].offset;
		if (INTEL_INFO(dev)->gen < 4)
			exec2_list[i].flags = EXEC_OBJECT_NEEDS_FENCE;
		else
			exec2_list[i].flags = 0;
	}

	exec2.buffers_ptr = args->buffers_ptr;
	exec2.buffer_count = args->buffer_count;
	exec2.batch_start_offset = args->batch_start_offset;
	exec2.batch_len = args->batch_len;
	exec2.DR1 = args->DR1;
	exec2.DR4 = args->DR4;
	exec2.num_cliprects = args->num_cliprects;
	exec2.cliprects_ptr = args->cliprects_ptr;
	exec2.flags = I915_EXEC_RENDER;
	i915_execbuffer2_set_context_id(exec2, 0);

	ret = kgrp_i915_gem_do_execbuffer(dev, data, file, &exec2, exec2_list);
	if (!ret) {
		struct drm_i915_gem_exec_object __user *user_exec_list =
			to_user_ptr(args->buffers_ptr);

		/* Copy the new buffer offsets back to the user's exec list. */
		for (i = 0; i < args->buffer_count; i++) {
			ret = __copy_to_user(&user_exec_list[i].offset,
					     &exec2_list[i].offset,
					     sizeof(user_exec_list[i].offset));
			if (ret) {
				ret = -EFAULT;
				KGRR_DRM_DEBUG("failed to copy %d exec entries "
					  "back to user (%d)\n",
					  args->buffer_count, ret);
				break;
			}
		}
	}

	drm_free_large(exec_list);
	drm_free_large(exec2_list);
	return ret;
}

int
kgrp_i915_gem_execbuffer2(struct drm_device *dev, void *data,
		     struct drm_file *file)
{
	struct drm_i915_gem_execbuffer2 *args = data;
	struct drm_i915_gem_exec_object2 *exec2_list = NULL;
	int ret;

	if (args->buffer_count < 1 ||
	    args->buffer_count > UINT_MAX / sizeof(*exec2_list)) {
		KGRR_DRM_DEBUG("execbuf2 with %d buffers\n", args->buffer_count);
		return -EINVAL;
	}

	if (args->rsvd2 != 0) {
		KGRR_DRM_DEBUG("dirty rvsd2 field\n");
		return -EINVAL;
	}

	exec2_list = kmalloc(sizeof(*exec2_list)*args->buffer_count,
			     GFP_TEMPORARY | __GFP_NOWARN | __GFP_NORETRY);
	if (exec2_list == NULL)
		exec2_list = drm_malloc_ab(sizeof(*exec2_list),
					   args->buffer_count);
	if (exec2_list == NULL) {
		KGRR_DRM_DEBUG("Failed to allocate exec list for %d buffers\n",
			  args->buffer_count);
		return -ENOMEM;
	}
	ret = copy_from_user(exec2_list,
			     to_user_ptr(args->buffers_ptr),
			     sizeof(*exec2_list) * args->buffer_count);
	if (ret != 0) {
		KGRR_DRM_DEBUG("copy %d exec entries failed %d\n",
			  args->buffer_count, ret);
		drm_free_large(exec2_list);
		return -EFAULT;
	}

	ret = kgrp_i915_gem_do_execbuffer(dev, data, file, args, exec2_list);
	if (!ret) {
		/* Copy the new buffer offsets back to the user's exec list. */
		struct drm_i915_gem_exec_object2 __user *user_exec_list =
				   to_user_ptr(args->buffers_ptr);
		int i;

		for (i = 0; i < args->buffer_count; i++) {
			ret = __copy_to_user(&user_exec_list[i].offset,
					     &exec2_list[i].offset,
					     sizeof(user_exec_list[i].offset));
			if (ret) {
				ret = -EFAULT;
				KGRR_DRM_DEBUG("failed to copy %d exec entries "
					  "back to user\n",
					  args->buffer_count);
				break;
			}
		}
	}

	drm_free_large(exec2_list);
	return ret;
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "i915_ggtt_view_normal", (void *)&kgre_i915_ggtt_view_normal,
	  "i915" },
	{ "drm_gem_object_free", (void *)&kgre_drm_gem_object_free, "drm" },
	{ "i915_gem_request_alloc", (void *)&kgre_i915_gem_request_alloc,
	  "i915" },
	{ "i915_gem_request_cancel", (void *)&kgre_i915_gem_request_cancel,
	  "i915" },
	{ "i915_gem_request_add_to_client",
	  (void *)&kgre_i915_gem_request_add_to_client, "i915" },
	{ "intel_runtime_pm_get", (void *)&kgre_intel_runtime_pm_get, "i915" },
	{ "intel_runtime_pm_put", (void *)&kgre_intel_runtime_pm_put, "i915" },
	{ "i915_needs_cmd_parser", (void *)&kgre_i915_needs_cmd_parser,
	  "i915" },
	{ "i915_gem_context_get", (void *)&kgre_i915_gem_context_get, "i915" },
	{ "i915_gem_context_free", (void *)&kgre_i915_gem_context_free,
	  "i915" },
	{ "i915_gem_evict_vm", (void *)&kgre_i915_gem_evict_vm, "i915" },
	{ "i915_gem_object_pin", (void *)&kgre_i915_gem_object_pin, "i915" },
	{ "i915_vma_unbind", (void *)&kgre_i915_vma_unbind, "i915" },
	{ "i915_gem_retire_requests_ring",
	  (void *)&kgre_i915_gem_retire_requests_ring, "i915" },
	{ "i915_gem_obj_ggtt_offset_view",
	  (void *)&kgre_i915_gem_obj_ggtt_offset_view, "i915" },
	{ "i915_gem_obj_offset", (void *)&kgre_i915_gem_obj_offset, "i915" },
	{ "i915_mutex_lock_interruptible",
	  (void *)&kgre_i915_mutex_lock_interruptible, "i915" },
	{ "i915_gem_object_ggtt_unpin_view",
	  (void *)&kgre_i915_gem_object_ggtt_unpin_view, "i915" },
	{ "i915_gem_object_get_fence", (void *)&kgre_i915_gem_object_get_fence,
	  "i915" },
	{ "i915_gem_object_pin_fence", (void *)&kgre_i915_gem_object_pin_fence,
	  "i915" },
	{ "i915_gem_obj_to_vma", (void *)&kgre_i915_gem_obj_to_vma, "i915" },
	{ "i915_gem_obj_lookup_or_create_vma",
	  (void *)&kgre_i915_gem_obj_lookup_or_create_vma, "i915" },
	{ "i915_gem_execbuffer_unreserve_vma",
	  (void *)&kgre_i915_gem_execbuffer_unreserve_vma, "i915" },
	{ "i915_gem_execbuffer_relocate_entry",
	  (void *)&kgre_i915_gem_execbuffer_relocate_entry, "i915" },
	{ "i915_gem_batch_pool_get", (void *)&kgre_i915_gem_batch_pool_get,
	  "i915" },
	{ "intel_ring_initialized", (void *)&kgre_intel_ring_initialized,
	  "i915" },
	{ "intel_lr_context_deferred_alloc",
	  (void *)&kgre_intel_lr_context_deferred_alloc, "i915" },
};

int kgr_bsc1173663_gem_execbuffer_resolve_kallsyms(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}

#endif
