/*
 * bsc1173663_gem_gtt.c
 *
 * Fix for CVE-2019-0155, bsc#1173663 (i915_gem_gtt.c part)
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)
#include <linux/seq_file.h>
#include <linux/stop_machine.h>

#include "bsc1173663_common.h"
#include "../kallsyms_relocs.h"

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.h */
static inline uint32_t i915_pte_index(uint64_t address, uint32_t pde_shift)
{
	const uint32_t mask = NUM_PTE(pde_shift) - 1;

	return (address >> PAGE_SHIFT) & mask;
}

static inline uint32_t i915_pde_index(uint64_t addr, uint32_t shift)
{
	return (addr >> shift) & I915_PDE_MASK;
}

#define gen8_for_each_pml4e(pdp, pml4, start, length, temp, iter)	\
	for (iter = gen8_pml4e_index(start);	\
	     length > 0 && iter < GEN8_PML4ES_PER_PML4 ? \
			(pdp = (pml4)->pdps[iter]), 1 : 0; \
	     iter++,				\
	     temp = ALIGN(start+1, 1ULL << GEN8_PML4E_SHIFT) - start,	\
	     temp = min(temp, length),					\
	     start += temp, length -= temp)

static inline uint32_t gen8_pte_index(uint64_t address)
{
	return i915_pte_index(address, GEN8_PDE_SHIFT);
}

static inline uint32_t gen8_pde_index(uint64_t address)
{
	return i915_pde_index(address, GEN8_PDE_SHIFT);
}

static inline uint32_t gen8_pdpe_index(uint64_t address)
{
	return (address >> GEN8_PDPE_SHIFT) & GEN8_PDPE_MASK;
}

static inline uint32_t gen8_pml4e_index(uint64_t address)
{
	return (address >> GEN8_PML4E_SHIFT) & GEN8_PML4E_MASK;
}

/* klp-ccp: from drivers/gpu/drm/i915/i915_drv.h */
#undef WARN_ON
#define WARN_ON(x) WARN((x), "WARN_ON(%s)", #x )

/* klp-ccp: from drivers/gpu/drm/i915/i915_trace.h */
KGRR_DEFINE_EVENT(i915_va, i915_va_alloc,
	     TP_PROTO(struct i915_address_space *vm, u64 start, u64 length, const char *name),
	     TP_ARGS(vm, start, length, name)
);


/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.c */
static void (*kgre_orig_gen8_ppgtt_insert_entries)(struct i915_address_space *vm,
				      struct sg_table *pages,
				      uint64_t start,
				      enum i915_cache_level cache_level,
				      u32 unused);

static void kgrp_gen8_ppgtt_insert_entries(struct i915_address_space *vm,
				      struct sg_table *pages,
				      uint64_t start,
				      enum i915_cache_level cache_level,
				      u32 flags);

static int kgrp_ppgtt_bind_vma(struct i915_vma *vma,
			  enum i915_cache_level cache_level,
			  u32 unused)
{
	u32 pte_flags = 0;

	/* Currently applicable only to VLV */
	if (vma->obj->gt_ro)
		pte_flags |= PTE_READ_ONLY;

	/*
	 * Fix CVE-2019-0155
	 *  +4 lines
	 * Double check that the callback is what it's expected to be.
	 * Note that the patched callchain should only get
	 * invoked when ->insert_entries == gen8_ppgtt_insert_entries.
	 */
	if (vma->vm->insert_entries == kgre_orig_gen8_ppgtt_insert_entries)
		kgrp_gen8_ppgtt_insert_entries(vma->vm, vma->obj->pages, vma->node.start,
					cache_level, pte_flags);
	else
		vma->vm->insert_entries(vma->vm, vma->obj->pages, vma->node.start,
				cache_level, pte_flags);

	return 0;
}

static gen8_pte_t kgrp_gen8_pte_encode(dma_addr_t addr,
				  enum i915_cache_level level,
/*
 * Fix CVE-2019-0155
 *  -1 line, +1 line
 */
				  bool valid, u32 flags)
{
	gen8_pte_t pte = valid ? _PAGE_PRESENT | _PAGE_RW : 0;
	pte |= addr;

	/*
	 * Fix CVE-2019-0155
	 *  +2 lines
	 */
	if (unlikely(flags & PTE_READ_ONLY))
		pte &= ~_PAGE_RW;

	switch (level) {
	case I915_CACHE_NONE:
		pte |= PPAT_UNCACHED_INDEX;
		break;
	case I915_CACHE_WT:
		pte |= PPAT_DISPLAY_ELLC_INDEX;
		break;
	default:
		pte |= PPAT_CACHED_INDEX;
		break;
	}

	return pte;
}

static void *kmap_page_dma(struct i915_page_dma *p)
{
	return kmap_atomic(p->page);
}

static void (*kgre_kunmap_page_dma)(struct drm_device *dev, void *vaddr);

#define kmap_px(px) kmap_page_dma(px_base(px))

static void
kgrp_gen8_ppgtt_insert_pte_entries(struct i915_address_space *vm,
			      struct i915_page_directory_pointer *pdp,
			      struct sg_page_iter *sg_iter,
			      uint64_t start,
/*
 * Fix CVE-2019-0155
 *  -1 line, +2 lines
 */
			      enum i915_cache_level cache_level,
			      u32 flags)
{
	struct i915_hw_ppgtt *ppgtt =
		container_of(vm, struct i915_hw_ppgtt, base);
	gen8_pte_t *pt_vaddr;
	unsigned pdpe = gen8_pdpe_index(start);
	unsigned pde = gen8_pde_index(start);
	unsigned pte = gen8_pte_index(start);

	pt_vaddr = NULL;

	while (__sg_page_iter_next(sg_iter)) {
		if (pt_vaddr == NULL) {
			struct i915_page_directory *pd = pdp->page_directory[pdpe];
			struct i915_page_table *pt = pd->page_table[pde];
			pt_vaddr = kmap_px(pt);
		}

		pt_vaddr[pte] =
			kgrp_gen8_pte_encode(sg_page_iter_dma_address(sg_iter),
					/*
					 * Fix CVE-2019-0155
					 *  -1 line, +1 line
					 */
					cache_level, true, flags);
		if (++pte == GEN8_PTES) {
			(*kgre_kunmap_page_dma)((ppgtt)->base.dev, (pt_vaddr));
			pt_vaddr = NULL;
			if (++pde == I915_PDES) {
				if (++pdpe == (((*kgre_i915).enable_ppgtt == 3) ? 512 : 4))
					break;
				pde = 0;
			}
			pte = 0;
		}
	}

	if (pt_vaddr)
		(*kgre_kunmap_page_dma)((ppgtt)->base.dev, (pt_vaddr));
}

static void kgrp_gen8_ppgtt_insert_entries(struct i915_address_space *vm,
				      struct sg_table *pages,
				      uint64_t start,
				      enum i915_cache_level cache_level,
/*
 * Fix CVE-2019-0155
 *  -1 line, +1 line
 */
				      u32 flags)
{
	struct i915_hw_ppgtt *ppgtt =
		container_of(vm, struct i915_hw_ppgtt, base);
	struct sg_page_iter sg_iter;

	__sg_page_iter_start(&sg_iter, pages->sgl, sg_nents(pages->sgl), 0);

	if (!((*kgre_i915).enable_ppgtt == 3)) {
		kgrp_gen8_ppgtt_insert_pte_entries(vm, &ppgtt->pdp, &sg_iter, start,
					      /*
					       * Fix CVE-2019-0155
					       *  -1 line, +1 line
					       */
					      cache_level, flags);
	} else {
		struct i915_page_directory_pointer *pdp;
		uint64_t templ4, pml4e;
		uint64_t length = (uint64_t)pages->orig_nents << PAGE_SHIFT;

		gen8_for_each_pml4e(pdp, &ppgtt->pml4, start, length, templ4, pml4e) {
			kgrp_gen8_ppgtt_insert_pte_entries(vm, pdp, &sg_iter,
						      /*
						       * Fix CVE-2019-0155
						       *  -1 line, +1 line
						       */
						      start, cache_level, flags);
		}
	}
}

static int (*kgre_orig_ppgtt_bind_vma)(struct i915_vma *vma,
			  enum i915_cache_level cache_level,
			  u32 unused);

/*
 * Clone of i915_vma_bind() called for execbuffer shadow objects on
 * gen8+ to be mapped as RO. This is to be extra conservative and not
 * livepatch gen8_ppgtt_insert_entries() directly. Instead the whole
 * callchain down to gen8_pte_encode() gets patched in order to make
 * sure that the PTE's RW bit really only gets cleared when invoked
 * from a patched caller, that is for shadowed execbuffer objects on
 * gen8+.
 */
int kgrp_i915_vma_bind(struct i915_vma *vma, enum i915_cache_level cache_level,
		  u32 flags)
{
	int ret;
	u32 bind_flags;

	if (WARN_ON(flags == 0))
		return -EINVAL;

	bind_flags = 0;
	if (flags & PIN_GLOBAL)
		bind_flags |= GLOBAL_BIND;
	if (flags & PIN_USER)
		bind_flags |= LOCAL_BIND;

	if (flags & PIN_UPDATE)
		bind_flags |= vma->bound;
	else
		bind_flags &= ~vma->bound;

	if (bind_flags == 0)
		return 0;

	if (vma->bound == 0 && vma->vm->allocate_va_range) {
		kgrr_trace_i915_va_alloc(vma->vm,
				    vma->node.start,
				    vma->node.size,
				    VM_TO_TRACE_NAME(vma->vm));

		/* XXX: i915_vma_pin() will fix this +- hack */
		vma->pin_count++;
		ret = vma->vm->allocate_va_range(vma->vm,
						 vma->node.start,
						 vma->node.size);
		vma->pin_count--;
		if (ret)
			return ret;
	}

	/*
	 * Fix CVE-2019-0155
	 *  +3 lines
	 */
	if (vma->vm->bind_vma == kgre_orig_ppgtt_bind_vma)
		ret = kgrp_ppgtt_bind_vma(vma, cache_level, bind_flags);
	else
		ret = vma->vm->bind_vma(vma, cache_level, bind_flags);
	if (ret)
		return ret;

	vma->bound |= bind_flags;

	return 0;
}

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "__tracepoint_i915_va_alloc",
	  (void *)&kgre___tracepoint_i915_va_alloc, "i915" },
	{ "kunmap_page_dma", (void *)&kgre_kunmap_page_dma, "i915" },
	{ "ppgtt_bind_vma", (void *)&kgre_orig_ppgtt_bind_vma, "i915" },
	{ "gen8_ppgtt_insert_entries",
	  (void *)&kgre_orig_gen8_ppgtt_insert_entries, "i915" },
};

int kgr_bsc1173663_gem_gtt_resolve_kallsyms(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}

#endif
